/* ---------------------------------------------------------------------------
   eibdrv_admin.c
   ---------------------------------------------------------------------------
   eibdrv_admin Version 0.2.1
   Copyright (C) 2002, Wolfgang Tumfart
                       Donaustrasse 104/9
                       A-2344 Maria Enzersdorf
                       Austria (Europe)
                       tumfart@auto.tuwien.ac.at

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA.
   --------------------------------------------------------------------------- */

/*  ---------------------------------------------------------------------------
    This program servers for output and change of internal parameter of
    the eibdrv driver like
    
       - mode, baudrate and serial port used,
       - size of buffers and their occupation
       
    and for visualization of internal statistical information about read
    and write actions delivered by the driver.
    
       Start: eibdrv_admin /dev/eibdrv   (Example)
    ---------------------------------------------------------------------------
 */   
/*  ---------------------------------------------------------------------------
    FormsLibrary is copyrighted by T.C. Zhao and Mark Overmars. It is not
    public domain, but may be used for running publically available free
    software.
    --------------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>

#include "forms.h"
#include "eibdrv.h"

#include "eibdrv_fl.h"


/*  Name of Panelfile for eibdrv_admin and number of message
    panels included.
 */
 
#define EIBDRV_ADMIN_PANELFILE "/sbin/eibdrv_admin.panel"
#define EIBDRV_ADMIN_PANELFILESIZE 13


#define EIBDRV_ADMIN_STATUS_NOTSET 0      // No status request sent yet
#define EIBDRV_ADMIN_STATUS_RESPONDED 1   // Status request msg acknowledged by BCU2
#define EIBDRV_ADMIN_STATUS_FAILED 2      // Status request msg ignored by BCU2

/*  Various color defines
 */

#define EIBDRV_ADMIN_COUNT_COL FL_PALEGREEN    // Size/Occupancy bar for values < 100%
#define EIBDRV_ADMIN_COUNT100_COL FL_TOMATO    // Size/Occupancy bar for value = 100%

#define EIBDRV_ADMIN_W1_COL FL_PALEGREEN       // Bar for 1-Writes and 0-Reads
#define EIBDRV_ADMIN_W2_COL FL_DARKCYAN        // Bar for 2-Writes and 1-Reads
#define EIBDRV_ADMIN_W3_COL FL_ORCHID          // Bar for 3-Writes and 2-Reads
#define EIBDRV_ADMIN_WFAILED_COL FL_TOMATO     // Bar for failed-Writes

#define EIBDRV_ADMIN_STATUS_RESPONDED_COL FL_PALEGREEN   // Box for Status msg acknowledged
#define EIBDRV_ADMIN_STATUS_FAILED_COL FL_TOMATO         // Box for Status msg ignored


/*  Main Form */

typedef struct 
   { FL_FORM *VisForm;
     void *vdata;
     char *cdata;
     long  ldata;

     FL_OBJECT *DeviceText;
     FL_OBJECT *ModeText;
     FL_OBJECT *BaudrateText;
     FL_OBJECT *PortText;

     FL_OBJECT *WProcSizeText;
     FL_OBJECT *RProcSizeText;
     FL_OBJECT *WProcOccText;
     FL_OBJECT *RProcOccText;
     FL_OBJECT *WProcOccBox;
     FL_OBJECT *WProcOccGroundBox;
     FL_OBJECT *RProcOccBox;
     FL_OBJECT *RProcOccGroundBox;
     FL_OBJECT *WMsgSizeText;
     FL_OBJECT *RMsgSizeText;
     FL_OBJECT *WMsgOccText;
     FL_OBJECT *WMsgOccBox;
     FL_OBJECT *WMsgOccGroundBox;
     
     FL_OBJECT *StatusBox;
     
     FL_OBJECT *TimeResetText;
     FL_OBJECT *TimeActualizeText;
     FL_OBJECT *TimeNextText;
     FL_OBJECT *ActIntervalCounter;
     FL_OBJECT *ActwoResetCheckbtn;
     FL_OBJECT *ActwResetCheckbtn;

     FL_OBJECT *WGroundBox;
     FL_OBJECT *WBox[4];
     FL_OBJECT *TotWText;
     FL_OBJECT *SecWText;
     FL_OBJECT *AvgWText;
     
     FL_OBJECT *RGroundBox;
     FL_OBJECT *RBox[3];
     FL_OBJECT *TotRText;
     FL_OBJECT *SecRText;
	
     FL_OBJECT *ResetStatBtn;
     FL_OBJECT *ActStatBtn;
     FL_OBJECT *ResizeBtn;
     FL_OBJECT *ResetProtBtn;
     FL_OBJECT *StatusBtn;
     FL_OBJECT *ExitBtn;
     
     FL_OBJECT *TimesTimer;
   } FD_VisForm;


/* Form for resizing process and message buffers */

typedef struct 
   { FL_FORM *VisForm;
     void *vdata;
     char *cdata;
     long  ldata;
     FL_OBJECT *WProcBufSizeEditfield;
     FL_OBJECT *RProcBufSizeEditfield;
     FL_OBJECT *WMsgBufSizeEditfield;
     FL_OBJECT *RMsgBufSizeEditfield;
     FL_OBJECT *OkBtn;
     FL_OBJECT *CancelBtn;
   } FD_ResizeBufferForm;


typedef struct
   { int mode;                         // Standard/Server mode of driver
     int symb_baudrate;                // symbolic constant of baudrate
     char *port;                       // serial port 
     unsigned long outbuf_size;        // outgoing messages (writes)
     unsigned long wpbuf_size;         // writing processes
     unsigned long rpbuf_size;         // reading processes
     unsigned long rpbuf_msgbuf_size;  // incoming messages (reads)
   } Eib_settings;
   

typedef struct 
   { FD_VisForm *MainForm;             // Main Form
     FD_ResizeBufferForm *ResizeBufferForm; // Resize buffers form
     int resbufform_visible;                // Resize buffers form shown/hidden
     PanelContent *Panels;             // all error and message panels
     int eib_fd;                       // file descriptor of eibdrv device
     int cleanup_code;                 // entry point in cleanup function
     Eib_settings *ftstation_settings; // settings of device
     Eib_statistics *ftstation_stat;   // internal statistics of device
     time_t stat_reset_time;           // time since last statistics reset
     time_t stat_get_time;             // time since last statistics actualization
     time_t stat_plan_time;            // time of next statistics actualization
     time_t act_interval;              // statistics actualization interval
   } Eib_bcu2_admin;


int baudrates[18][2]={ {B50,50},       // Table of symbolic baudrates and their integer 
                       {B75,75},       // value
                       {B110,110},
                       {B134,134},
                       {B150,150},
                       {B200,200},
                       {B300,300},
                       {B600,600},
                       {B1200,1200},
                       {B1800,1800},
                       {B2400,2400},
                       {B4800,4800}, 
                       {B9600,9600},
                       {B19200,19200},
                       {B38400,38400},
                       {B57600,57600},
                       {B115200,115200},
                       {B230400,230400}};


/*  --------------------------------------------------------------------------- 
    Various function prototypes.
    ---------------------------------------------------------------------------
*/

FD_VisForm *create_form_MainForm(void);
FD_ResizeBufferForm *create_form_ResizeBufferForm(void);
void cleanup(void *arg,int btn);


/*  --------------------------------------------------------------------------- 
    Allocation and deallocation of central struct of administrator.
    ---------------------------------------------------------------------------
 */
 
Eib_bcu2_admin *getEib_bcu2_admin()
{  Eib_bcu2_admin *help=(Eib_bcu2_admin *)fl_calloc(1,sizeof(Eib_bcu2_admin));

   help->resbufform_visible=FL_EXT_INVISIBLE;
   help->Panels=NULL;
   help->cleanup_code=0;
   help->ftstation_settings=(Eib_settings *)fl_calloc(1,sizeof(Eib_settings));
   help->ftstation_settings->port=NULL;
   help->ftstation_stat=(Eib_statistics *)fl_calloc(1,sizeof(Eib_statistics));
   return help;
}

void returnEib_bcu2_admin(Eib_bcu2_admin *admin)
{  free(admin->MainForm);
   free(admin->ResizeBufferForm);
   returnPanelContentA(admin->Panels,EIBDRV_ADMIN_PANELFILESIZE);
   if (admin->ftstation_settings->port!=NULL) free(admin->ftstation_settings->port);
   free(admin->ftstation_settings);
   free(admin->ftstation_stat);
   free(admin);
}

 
/*  --------------------------------------------------------------------------- 
    Creation of all forms and callbacks of all visual elements.
    ---------------------------------------------------------------------------
 */

void actualize_MainForm_BufSiz(Eib_bcu2_admin *admin)
{  char *helpstring=(char *)malloc(50*sizeof(char));

   if (admin->ftstation_settings->mode==FT_SERVERMODE)
   {  sprintf(helpstring,"%li",admin->ftstation_settings->wpbuf_size);
      fl_set_object_label(admin->MainForm->WProcSizeText,helpstring);  

      sprintf(helpstring,"%li",admin->ftstation_settings->rpbuf_size);
      fl_set_object_label(admin->MainForm->RProcSizeText,helpstring);  
 
      sprintf(helpstring,"%li",admin->ftstation_settings->outbuf_size);
      fl_set_object_label(admin->MainForm->WMsgSizeText,helpstring);  

      sprintf(helpstring,"%li",admin->ftstation_settings->rpbuf_msgbuf_size);
      fl_set_object_label(admin->MainForm->RMsgSizeText,helpstring);  
   }
   free(helpstring);
}


void actualize_MainForm_Stat(Eib_bcu2_admin *admin)
{  char *helpstring=(char *)malloc(50*sizeof(char));
   Eib_settings *settings=admin->ftstation_settings;
   Eib_statistics *stat=admin->ftstation_stat;
   FD_VisForm *form=admin->MainForm;
   double diff_time;
   unsigned long rw_ges;
   double rw_pt;
   int lowerbound;
   int i;
   int avg_w1time;

   if (admin->ftstation_settings->mode==FT_SERVERMODE)
   {  sprintf(helpstring,"%li",stat->wpbuf_count);
      fl_set_object_label(form->WProcOccText,helpstring);

      sprintf(helpstring,"%li",stat->rpbuf_count);
      fl_set_object_label(form->RProcOccText,helpstring);

      sprintf(helpstring,"%li",stat->outbuf_count);
      fl_set_object_label(form->WMsgOccText,helpstring);
   
      fl_redraw_object(form->WProcOccGroundBox);
      if (stat->wpbuf_count==settings->wpbuf_size)
      {  form->WProcOccBox->w=96;
         fl_set_object_color(form->WProcOccBox,EIBDRV_ADMIN_COUNT100_COL,BOX_COL2);
      }
      else
      {  form->WProcOccBox->w=96.0*((float)stat->wpbuf_count/(float)settings->wpbuf_size);
         fl_set_object_color(form->WProcOccBox,EIBDRV_ADMIN_COUNT_COL,BOX_COL2);
      }
      fl_redraw_object(form->WProcOccBox);

      fl_redraw_object(form->RProcOccGroundBox);   
      if (stat->rpbuf_count==settings->rpbuf_size)
      {  form->RProcOccBox->w=96;
         fl_set_object_color(form->RProcOccBox,EIBDRV_ADMIN_COUNT100_COL,BOX_COL2);
      }
      else
      {  form->RProcOccBox->w=96.0*((float)stat->rpbuf_count/(float)settings->rpbuf_size);
         fl_set_object_color(form->RProcOccBox,EIBDRV_ADMIN_COUNT_COL,BOX_COL2);
      }
      fl_redraw_object(form->RProcOccBox);

      fl_redraw_object(form->WMsgOccGroundBox);
      if (stat->outbuf_count==settings->outbuf_size)
      {  form->WMsgOccBox->w=96;
         fl_set_object_color(form->WMsgOccBox,EIBDRV_ADMIN_COUNT100_COL,BOX_COL2);
      }
      else
      {  form->WMsgOccBox->w=96.0*((float)stat->outbuf_count/(float)settings->outbuf_size);
         fl_set_object_color(form->WMsgOccBox,EIBDRV_ADMIN_COUNT_COL,BOX_COL2);
      }
      fl_redraw_object(form->WMsgOccBox);
   }
   
   fl_redraw_object(form->WGroundBox);
   fl_redraw_object(form->RGroundBox); 

   if (stat->overflow==1)
   {  fl_set_object_label(form->TotWText,"");
      fl_set_object_label(form->TotRText,"");
      fl_set_object_label(form->SecWText,"");
      fl_set_object_label(form->SecRText,"");
      fl_set_object_label(form->AvgWText,"");
      free(helpstring);      
      return;
   }      

   diff_time=difftime(admin->stat_get_time,admin->stat_reset_time);
   sprintf(helpstring,"%li - %li - %li - %li",stat->write[1],stat->write[2],stat->write[3],stat->write[0]);
   fl_set_object_label(form->TotWText,helpstring);
   sprintf(helpstring,"%li - %li - %li",stat->read[0],stat->read[1],stat->read[2]);
   fl_set_object_label(form->TotRText,helpstring);

   if (diff_time>0)
   {  sprintf(helpstring,"%3.2f - %3.2f - %3.2f - %3.2f",stat->write[1]/diff_time,stat->write[2]/diff_time,
         stat->write[3]/diff_time,stat->write[0]/diff_time);
      fl_set_object_label(form->SecWText,helpstring);
      sprintf(helpstring,"%3.2f - %3.2f - %3.2f",stat->read[0]/diff_time,stat->read[1]/diff_time,
         stat->read[2]/diff_time);
      fl_set_object_label(form->SecRText,helpstring);
   }
   else
   {  fl_set_object_label(form->SecWText,"");
      fl_set_object_label(form->SecRText,"");
   }
  
   rw_ges=0;
   for (i=0;i<4;i++) rw_ges+=stat->write[i];
   rw_pt=273.0/rw_ges;
   lowerbound=41;
   for (i=1;i<5;i++)
   {  if (stat->write[i%4]>0)
      {  form->WBox[i%4]->x=lowerbound;
         form->WBox[i%4]->w=314-lowerbound;
         fl_redraw_object(form->WBox[i%4]);
         lowerbound+=(int)(stat->write[i%4]*rw_pt);
      }
   }
      
   rw_ges=0;
   for (i=0;i<3;i++) rw_ges+=stat->read[i];
   rw_pt=273.0/rw_ges;
   lowerbound=376;
   for (i=0;i<3;i++)
   {  if (stat->read[i]>0)
      {  form->RBox[i]->x=lowerbound;
         form->RBox[i]->w=649-lowerbound;
         fl_redraw_object(form->RBox[i]);
         lowerbound+=(int)(stat->read[i]*rw_pt);
      }
   }

   if (stat->write[1]>0)   
   {  avg_w1time=((stat->total_w1time.tv_sec+stat->total_w1time.tv_usec/1000000.0)*1000)/stat->write[1];
      if (avg_w1time<1000)
      {  sprintf(helpstring,"%i ms",avg_w1time);
         fl_set_object_label(form->AvgWText,helpstring);
      }
      else
         fl_set_object_label(form->AvgWText,"");
   }
   else
      fl_set_object_label(form->AvgWText,"");

   free(helpstring);
}


void actualize_MainForm_Times(Eib_bcu2_admin *admin,time_t act_time)
{  unsigned long diff_stat_reset,diff_stat_get,diff_stat_plan;
   char timestring[9];

   diff_stat_reset=(unsigned long int)difftime(act_time,admin->stat_reset_time);
   if (diff_stat_reset<=359999)
   {  sprintf(&timestring[0],"%02li:%02li:%02li",diff_stat_reset/3600,
         (diff_stat_reset%3600)/60,diff_stat_reset%60);
      fl_set_object_label(admin->MainForm->TimeResetText,&timestring[0]);
   }
   else
      fl_set_object_label(admin->MainForm->TimeResetText,"");

   diff_stat_get=(unsigned long int)difftime(act_time,admin->stat_get_time);
   if (diff_stat_get<=359999)
   {  sprintf(&timestring[0],"%02li:%02li:%02li",diff_stat_get/3600,
         (diff_stat_get%3600)/60,diff_stat_get%60);
      fl_set_object_label(admin->MainForm->TimeActualizeText,&timestring[0]);
   }
   else
      fl_set_object_label(admin->MainForm->TimeActualizeText,"");
      
   diff_stat_plan=(unsigned long int)difftime(admin->stat_plan_time,act_time);
   if (diff_stat_plan<=359999)
   {  sprintf(&timestring[0],"%02li:%02li:%02li",diff_stat_plan/3600,
         (diff_stat_plan%3600)/60,diff_stat_plan%60);
      fl_set_object_label(admin->MainForm->TimeNextText,&timestring[0]);
   }
   else
      fl_set_object_label(admin->MainForm->TimeNextText,"");
}   


void actualize_MainForm_Status(Eib_bcu2_admin *admin,int status)
{  switch(status)
   {  case EIBDRV_ADMIN_STATUS_NOTSET:
              fl_set_object_label(admin->MainForm->StatusBox,"");
              fl_set_object_color(admin->MainForm->StatusBox,BOX_COL1,BOX_COL2);
              break;
      case EIBDRV_ADMIN_STATUS_RESPONDED:
              fl_set_object_label(admin->MainForm->StatusBox,"RESPONDED");
              fl_set_object_color(admin->MainForm->StatusBox,
                 EIBDRV_ADMIN_STATUS_RESPONDED_COL,BOX_COL2);
              break;
      case EIBDRV_ADMIN_STATUS_FAILED:
              fl_set_object_label(admin->MainForm->StatusBox,"FAILED");
              fl_set_object_color(admin->MainForm->StatusBox,
                 EIBDRV_ADMIN_STATUS_FAILED_COL,BOX_COL2);
              break;
   }              
}


void panel_cb(void *arg,int btn)
{  ;  }

   
void Null_cb(FL_OBJECT *obj,long data)
{  ;  }


void ResizeBufferPanel_cb(void *arg,int btn)
{  Eib_bcu2_admin *admin=(Eib_bcu2_admin *)arg;
   
   fl_activate_form(admin->MainForm->VisForm);
}

   
void ActIntervalCounter_cb(FL_OBJECT *obj, long data)
{  Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
   
   admin->act_interval=fl_get_counter_value(obj)*60;
   admin->stat_plan_time=time(NULL)+admin->act_interval;
}


void ResetProtBtn_cb(FL_OBJECT *obj, long data)
{  int result;
   Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
   
   if (ioctl(admin->eib_fd,FT_RESET)<0)
   {  switch(result=errno)
      {  case ENORESET: fl_ext_panel1(admin->Panels+9,admin,panel_cb);
                        break;
      }
   }
   else
      fl_ext_panel1(admin->Panels+10,admin,panel_cb);
}


void StatusBtn_cb(FL_OBJECT *obj, long data)
{  int result;
   Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
   
   if (ioctl(admin->eib_fd,FT_RESP_STATUS)<0)
   {  switch(result=errno)
      {  case ENORESPSTATUS: fl_freeze_form(admin->MainForm->VisForm);
                             actualize_MainForm_Status(admin,EIBDRV_ADMIN_STATUS_FAILED);
                             fl_unfreeze_form(admin->MainForm->VisForm);
                             fl_ext_panel1(admin->Panels+7,admin,panel_cb);
                             break;
      }
   }
   else
   {  fl_freeze_form(admin->MainForm->VisForm);
      actualize_MainForm_Status(admin,EIBDRV_ADMIN_STATUS_RESPONDED);
      fl_ext_panel1(admin->Panels+6,admin,panel_cb);
      fl_unfreeze_form(admin->MainForm->VisForm);
   } 
}


void ExitBtn_cb(FL_OBJECT *obj, long data)
{  Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
   
   fl_hide_form(admin->MainForm->VisForm);
   fl_free_form(admin->MainForm->VisForm);
   fl_free_form(admin->ResizeBufferForm->VisForm);
   close(admin->eib_fd);
   returnEib_bcu2_admin(admin);
   fl_finish();
   exit(0);
}


void ResizeBtn_cb(FL_OBJECT *obj, long data)
{  char *helpstring=(char *)malloc(50*sizeof(char));
   Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
      
   sprintf(helpstring,"%li",admin->ftstation_settings->wpbuf_size);
   fl_set_input(admin->ResizeBufferForm->WProcBufSizeEditfield,helpstring);  

   sprintf(helpstring,"%li",admin->ftstation_settings->rpbuf_size);
   fl_set_input(admin->ResizeBufferForm->RProcBufSizeEditfield,helpstring);  

   sprintf(helpstring,"%li",admin->ftstation_settings->outbuf_size);
   fl_set_input(admin->ResizeBufferForm->WMsgBufSizeEditfield,helpstring);  

   sprintf(helpstring,"%li",admin->ftstation_settings->rpbuf_msgbuf_size);
   fl_set_input(admin->ResizeBufferForm->RMsgBufSizeEditfield,helpstring);  
   
   fl_deactivate_form(admin->MainForm->VisForm);
   fl_show_form(admin->ResizeBufferForm->VisForm,FL_PLACE_CENTER,FL_TRANSIENT,"eibdrv_admin");
   admin->resbufform_visible=FL_EXT_VISIBLE;
   
   free(helpstring);
}


void OkBtn_cb(FL_OBJECT *obj, long data)
{  int result_wpbuf=0;
   int result_rpbuf=0;
   int result_outbuf=0;
   int result_rpbuf_msgbuf=0;
   int deact_flag=0;
   unsigned long new_size;
   Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
    
   if (strcmp(fl_get_input(admin->ResizeBufferForm->WProcBufSizeEditfield),"")!=0)
   {  new_size=(unsigned long)atol(fl_get_input(admin->ResizeBufferForm->WProcBufSizeEditfield));
      if (ioctl(admin->eib_fd,FT_SET_WPBUF_SIZE,new_size)<0)
         result_wpbuf=errno;
      else
         admin->ftstation_settings->wpbuf_size=new_size;
   }
   if (strcmp(fl_get_input(admin->ResizeBufferForm->RProcBufSizeEditfield),"")!=0)
   {  new_size=(unsigned long)atol(fl_get_input(admin->ResizeBufferForm->RProcBufSizeEditfield));
      if (ioctl(admin->eib_fd,FT_SET_RPBUF_SIZE,new_size)<0)
         result_rpbuf=errno;
      else
         admin->ftstation_settings->rpbuf_size=new_size;
   }
   if (strcmp(fl_get_input(admin->ResizeBufferForm->WMsgBufSizeEditfield),"")!=0)
   {  new_size=(unsigned long)atol(fl_get_input(admin->ResizeBufferForm->WMsgBufSizeEditfield));
      if (ioctl(admin->eib_fd,FT_SET_OUTBUF_SIZE,new_size)<0)
         result_outbuf=errno;
      else
         admin->ftstation_settings->outbuf_size=new_size;
   }
   if (strcmp(fl_get_input(admin->ResizeBufferForm->RMsgBufSizeEditfield),"")!=0)
   {  new_size=(unsigned long)atol(fl_get_input(admin->ResizeBufferForm->RMsgBufSizeEditfield));
      if (ioctl(admin->eib_fd,FT_SET_RPBUF_MSGSIZE,new_size)<0)
         result_rpbuf_msgbuf=errno;
      else
         admin->ftstation_settings->rpbuf_msgbuf_size=new_size;
   }
            
   fl_freeze_form(admin->MainForm->VisForm);
   actualize_MainForm_BufSiz(admin);
   fl_unfreeze_form(admin->MainForm->VisForm);
         
   if ((result_wpbuf==ENOMEM) || (result_rpbuf==ENOMEM) || 
       (result_outbuf==ENOMEM) || (result_rpbuf_msgbuf==ENOMEM))
   {  fl_ext_panel1(admin->Panels+11,admin,ResizeBufferPanel_cb);
      deact_flag=1;
   }

   if ((result_wpbuf==ESIZE) || (result_rpbuf==ESIZE) || 
       (result_outbuf==ESIZE) || (result_rpbuf_msgbuf==ESIZE))
   {  fl_ext_panel1(admin->Panels+12,admin,ResizeBufferPanel_cb);
      deact_flag=1;
   }
      
   fl_hide_form(admin->ResizeBufferForm->VisForm);
   if (deact_flag==0)
      fl_activate_form(admin->MainForm->VisForm);
   time(&admin->stat_plan_time);
}


void CancelBtn_cb(FL_OBJECT *obj, long data)
{  Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
      
   fl_hide_form(admin->ResizeBufferForm->VisForm);
   admin->resbufform_visible=FL_EXT_INVISIBLE;
   fl_activate_form(admin->MainForm->VisForm);
}


void TimesTimer_cb(FL_OBJECT *obj,long use_type)

/*  The argument use_type determines, whether the statistics record has to be
    initialized after reading it (FT_GETRESET_STAT) or not (FT_GET_STAT).
    
      0 ... The checkbuttons ActwResetCheckBtn and ActwoResetCheckBtn 
            determine the ioctl call.
      1 ... The ioctl is called in FT_GETRESET_STAT mode.
      2 ... The ioctl is called in FT_GET_STAT mode.
 */
 
{  Eib_bcu2_admin *admin=(Eib_bcu2_admin *)obj->form->u_vdata;
   time_t act_time;
   int result;
   unsigned int cmd;
      
   time(&act_time);
   actualize_MainForm_Times(admin,act_time);
   
   switch(use_type)
   {  case 0: if (fl_get_button(admin->MainForm->ActwoResetCheckbtn)==1)
                 cmd=FT_GET_STAT;
              else
                 cmd=FT_GETRESET_STAT;
              break;
      case 1: cmd=FT_GETRESET_STAT; 
              admin->stat_plan_time=act_time;
              break;
      case 2: cmd=FT_GET_STAT; 
              admin->stat_plan_time=act_time;
              break;
   }
      
   if (admin->stat_plan_time<=act_time)   
   {  fl_freeze_form(admin->MainForm->VisForm);
      if (ioctl(admin->eib_fd,cmd,admin->ftstation_stat)<0)
      {  switch(result=errno)
         {  case EFAULT: fl_ext_panel1(admin->Panels+5,admin,cleanup);
                         break;
         }
      }
   
      if (ioctl(admin->eib_fd,FT_RESP_STATUS)<0)
      {  switch(result=errno)
         {  case ENORESPSTATUS: actualize_MainForm_Status(admin,EIBDRV_ADMIN_STATUS_FAILED);
                                break;
         }
      }
      else
         actualize_MainForm_Status(admin,EIBDRV_ADMIN_STATUS_RESPONDED);

      admin->stat_get_time=admin->stat_plan_time;
      admin->stat_plan_time+=admin->act_interval;
      actualize_MainForm_BufSiz(admin);
      actualize_MainForm_Stat(admin);

      if (cmd==FT_GETRESET_STAT)
         admin->stat_reset_time=admin->stat_get_time;
   
      fl_unfreeze_form(admin->MainForm->VisForm);
   }
   
   if (use_type==0)  // Timer   
      fl_set_timer(obj,1);
}


FD_VisForm *create_form_MainForm(void)
{  int i;
   FL_OBJECT *obj;
   FD_VisForm *fdui = (FD_VisForm *) fl_calloc(1, sizeof(*fdui));

   fdui->VisForm = fl_bgn_form(FL_NO_BOX, 800, 475);

   fl_ext_formobjects_add_box(FL_UP_BOX,0,0,800,475,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_SouthEast);
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,20,420,115,"Settings",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,30,55,20,"Device:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,55,55,20,"Mode:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,80,55,20,"Baudrate:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,105,55,20,"Port:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->DeviceText,FL_BORDER_BOX,105,30,315,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&fdui->ModeText,FL_BORDER_BOX,105,55,315,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&fdui->BaudrateText,FL_BORDER_BOX,105,80,315,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&fdui->PortText,FL_BORDER_BOX,105,105,315,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,460,20,130,80,"Protocol Status",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,470,35,110,20,"Controller to BCU2",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_box2(&fdui->StatusBox,FL_BORDER_BOX,475,65,100,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      

   fl_ext_formobjects_add_box(FL_EMBOSSED_BOX,610,20,170,80,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,620,25,150,30,"eibdrv_admin",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lcolor(obj,FL_BLUE);
      fl_set_object_lsize(obj,FL_LARGE_SIZE);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_EMBOSSED_STYLE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,660,50,110,20,"Version 0.2.1",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);         
      fl_set_object_lsize(obj,FL_NORMAL_SIZE);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_BOLD_STYLE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,650,70,120,20,"(C) Wolfgang Tumfart",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);         
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,155,420,65,"Process Buffer",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,165,45,20,"WRITE:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,190,45,20,"READ:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,165,30,20,"Size:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,190,30,20,"Size:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->WProcSizeText,FL_BORDER_BOX,115,165,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&fdui->RProcSizeText,FL_BORDER_BOX,115,190,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,165,165,55,20,"Occupied:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,165,190,55,20,"Occupied:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->WProcOccText,FL_BORDER_BOX,220,165,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&fdui->RProcOccText,FL_BORDER_BOX,220,190,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,270,165,25,20,"0 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,270,190,25,20,"0 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,395,165,40,20,"100 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,395,190,40,20,"100 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_box2(&fdui->WProcOccGroundBox,FL_DOWN_BOX,295,165,100,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_box2(&fdui->RProcOccGroundBox,FL_DOWN_BOX,295,190,100,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_box2(&fdui->WProcOccBox,FL_FLAT_BOX,297,167,0,16,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_box2(&fdui->RProcOccBox,FL_FLAT_BOX,297,192,0,16,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,240,420,65,"Message Buffer",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,250,45,20,"WRITE:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,275,45,20,"READ:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,250,30,20,"Size:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,275,30,20,"Size:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->WMsgSizeText,FL_BORDER_BOX,115,250,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&fdui->RMsgSizeText,FL_BORDER_BOX,115,275,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,165,250,55,20,"Occupied:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->WMsgOccText,FL_BORDER_BOX,220,250,40,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,270,250,25,20,"0 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,395,250,40,20,"100 %",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_box2(&fdui->WMsgOccGroundBox,FL_DOWN_BOX,295,250,100,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_box2(&fdui->WMsgOccBox,FL_FLAT_BOX,297,252,0,16,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,460,120,210,185,"Statistics Actualization",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,130,100,20,"Last Reset:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,155,100,20,"Last Actualization:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,180,100,20,"Next Actualization:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->TimeResetText,FL_BORDER_BOX,580,130,70,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_text(&fdui->TimeActualizeText,FL_BORDER_BOX,580,155,70,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_text(&fdui->TimeNextText,FL_BORDER_BOX,580,180,70,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_counter(&fdui->ActIntervalCounter,FL_NORMAL_COUNTER,
         475,210,175,20,"Actualization Intervall (min)",FL_ALIGN_BOTTOM,
         1.0,60.0,1.0,10.0,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
      fl_set_object_callback(fdui->ActIntervalCounter,ActIntervalCounter_cb,0);
      fl_set_counter_value(fdui->ActIntervalCounter,1);
   fl_bgn_group();
   fl_ext_formobjects_add_checkbtn(&fdui->ActwoResetCheckbtn,FL_RADIO_BUTTON,
         475,250,165,20,"Actualization without Reset",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
      fl_set_object_callback(fdui->ActwoResetCheckbtn,Null_cb,0);
      fl_set_button(fdui->ActwoResetCheckbtn,1);
   fl_ext_formobjects_add_checkbtn(&fdui->ActwResetCheckbtn,FL_RADIO_BUTTON,
         475,270,165,20,"Actualization with Reset",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
      fl_set_object_callback(fdui->ActwResetCheckbtn,Null_cb,0);
   fl_end_group();
   
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,325,315,130,"Write (Messages to EIB system)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
   fl_ext_formobjects_add_box2(&fdui->WGroundBox,FL_BORDER_BOX,40,355,275,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   for (i=1;i<5;i++)
      fl_ext_formobjects_add_box2(&fdui->WBox[i%4],FL_FLAT_BOX,41,356,0,18,"",
            FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_set_object_color(fdui->WBox[1],EIBDRV_ADMIN_W1_COL,BOX_COL2);
   fl_set_object_color(fdui->WBox[2],EIBDRV_ADMIN_W2_COL,BOX_COL2);
   fl_set_object_color(fdui->WBox[3],EIBDRV_ADMIN_W3_COL,BOX_COL2);
   fl_set_object_color(fdui->WBox[0],EIBDRV_ADMIN_WFAILED_COL,BOX_COL2);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,335,165,20,"1 - 2 - 3 attempts - failed",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lcolor(obj,FL_RIGHT_BCOL);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->TotWText,FL_FLAT_BOX,40,375,205,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(fdui->TotWText,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,260,375,60,20,"total writes",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->SecWText,FL_FLAT_BOX,40,390,205,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(fdui->SecWText,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,260,390,60,20,"writes / sec",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,415,140,15,"Average total time for write",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,430,170,15,"with success in first attempt (ms):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->AvgWText,FL_BORDER_BOX,205,425,110,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
         
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,355,325,315,130,"Read (Messages from EIB system)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
   fl_ext_formobjects_add_box2(&fdui->RGroundBox,FL_BORDER_BOX,375,355,275,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   for (i=0;i<3;i++)
      fl_ext_formobjects_add_box2(&fdui->RBox[i],FL_FLAT_BOX,376,356,0,18,"",
            FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_set_object_color(fdui->RBox[0],EIBDRV_ADMIN_W1_COL,BOX_COL2);
   fl_set_object_color(fdui->RBox[1],EIBDRV_ADMIN_W2_COL,BOX_COL2);
   fl_set_object_color(fdui->RBox[2],EIBDRV_ADMIN_W3_COL,BOX_COL2);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,375,335,130,20,"0 - 1 - 2  retranmissions",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);                  
      fl_set_object_lcolor(obj,FL_RIGHT_BCOL);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->TotRText,FL_FLAT_BOX,375,375,205,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);                  
      fl_set_object_lalign(fdui->TotRText,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,595,375,60,20,"total reads",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);                  
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&fdui->SecRText,FL_FLAT_BOX,375,390,205,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);                  
      fl_set_object_lalign(fdui->SecRText,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,595,390,60,20,"reads / sec",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);

   fl_ext_formobjects_add_btn(&fdui->ResetStatBtn,FL_NORMAL_BUTTON,690,120,90,25,"Reset Statist.",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_callback(fdui->ResetStatBtn,TimesTimer_cb,1);
   fl_ext_formobjects_add_btn(&fdui->ActStatBtn,FL_NORMAL_BUTTON,690,155,90,25,"Actualize Statist.",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_callback(fdui->ActStatBtn,TimesTimer_cb,2);
   fl_ext_formobjects_add_btn(&fdui->ResizeBtn,FL_NORMAL_BUTTON,690,280,90,25,"Resize Buffers",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);     
      fl_set_object_callback(fdui->ResizeBtn,ResizeBtn_cb,0);
   fl_ext_formobjects_add_btn(&fdui->ResetProtBtn,FL_NORMAL_BUTTON,690,360,90,25,"Reset Protocol",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->ResetProtBtn,ResetProtBtn_cb,0);
   fl_ext_formobjects_add_btn(&fdui->StatusBtn,FL_NORMAL_BUTTON,690,395,90,25,"Request Status",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_callback(fdui->StatusBtn,StatusBtn_cb,0);
   fl_ext_formobjects_add_btn(&fdui->ExitBtn,FL_NORMAL_BUTTON,690,430,90,25,"Exit",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
      fl_set_object_callback(fdui->ExitBtn,ExitBtn_cb,0);

   fdui->TimesTimer=fl_add_timer(FL_HIDDEN_TIMER,1,1,0,0,"");
      fl_set_object_callback(fdui->TimesTimer,TimesTimer_cb,0);

   fl_end_form();
   fdui->VisForm->fdui = fdui;

   return fdui;
}


FD_ResizeBufferForm *create_form_ResizeBufferForm(void)
{  FL_OBJECT *obj;
   FD_ResizeBufferForm *fdui = (FD_ResizeBufferForm *) fl_calloc(1, sizeof(*fdui));

   fdui->VisForm = fl_bgn_form(FL_NO_BOX, 340, 190);
   fl_ext_formobjects_add_box(FL_UP_BOX,0,0,340,190,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,20,190,65,"Process Buffer Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,30,55,20,"WRITE:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,55,55,20,"READ:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_inp(&fdui->WProcBufSizeEditfield,FL_NORMAL_INPUT,
         115,30,80,20,"Size:",0,FL_RETURN_END_CHANGED,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);         
      fl_set_object_callback(fdui->WProcBufSizeEditfield,Null_cb,0);
      fl_set_input_filter(fdui->WProcBufSizeEditfield,dec_input_filter);
   fl_ext_formobjects_add_inp(&fdui->RProcBufSizeEditfield,FL_NORMAL_INPUT,
         115,55,80,20,"Size:",0,FL_RETURN_END_CHANGED,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
      fl_set_object_callback(fdui->RProcBufSizeEditfield,Null_cb,0);
      fl_set_input_filter(fdui->RProcBufSizeEditfield,dec_input_filter);

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,105,190,65,"Message Buffer Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,115,55,20,"WRITE:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,140,55,20,"READ:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);      
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
   fl_ext_formobjects_add_inp(&fdui->WMsgBufSizeEditfield,FL_NORMAL_INPUT,
         115,115,80,20,"Size:",0,FL_RETURN_END_CHANGED,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);         
      fl_set_object_callback(fdui->WMsgBufSizeEditfield,Null_cb,0);
      fl_set_input_filter(fdui->WMsgBufSizeEditfield,dec_input_filter);
   fl_ext_formobjects_add_inp(&fdui->RMsgBufSizeEditfield,FL_NORMAL_INPUT,
         115,140,80,20,"Size:",0,FL_RETURN_END_CHANGED,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
      fl_set_object_callback(fdui->RMsgBufSizeEditfield,Null_cb,0);
      fl_set_input_filter(fdui->RMsgBufSizeEditfield,dec_input_filter);

   fl_ext_formobjects_add_btn(&fdui->OkBtn,FL_NORMAL_BUTTON,230,110,90,25,"OK",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_callback(fdui->OkBtn,OkBtn_cb,0);
   fl_ext_formobjects_add_btn(&fdui->CancelBtn,FL_NORMAL_BUTTON,230,145,90,25,"Cancel",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);               
      fl_set_object_callback(fdui->CancelBtn,CancelBtn_cb,0);
   
   fl_end_form();
   fdui->VisForm->fdui = fdui;
   return fdui;
}


/*  --------------------------------------------------------------------------- 

    ---------------------------------------------------------------------------
 */

void cleanup(void *arg,int btn)
{  
   /*  Schliessen offener Dateien und Beenden der visuellen Forms, Rueckgabe 
       der zentralen administrator Struktur.
    */
   
   Eib_bcu2_admin *admin=(Eib_bcu2_admin *)arg;
   
   switch(admin->cleanup_code)
   {  case 3:  fl_hide_form(admin->MainForm->VisForm);
               fl_free_form(admin->MainForm->VisForm);
               fl_free_form(admin->ResizeBufferForm->VisForm);
      case 2:  close(admin->eib_fd);
      case 1:  returnEib_bcu2_admin(admin);
               fl_finish();
               exit(3);
   }
}


int main(int argc,char *argv[])
{  int len,i,result;
   char BaudrateText[10];
   Eib_bcu2_admin *admin;
   
   admin=getEib_bcu2_admin();

   if (read_PanelContentA(EIBDRV_ADMIN_PANELFILE,&admin->Panels)!=EIBDRV_ADMIN_PANELFILESIZE)
   {  printf("eibdrv_admin: Panelfile %s not found!\n",EIBDRV_ADMIN_PANELFILE);
      exit(2);
   }
   
   if (argc!=2)
   {  printf("usage: eibdrv_admin device\n");
      exit(1);
   }

   fl_set_icm_color(FL_INACTIVE_COL,200,200,200);
   fl_initialize(&argc,argv,"eibdrv_admin",0,0);
   fl_set_border_width(2);

   admin->cleanup_code=1;
   
   if ((admin->eib_fd=open(argv[1],O_RDWR|O_NONBLOCK))<0)
   {  switch(result=errno)
      {  case ENOENT:  fl_ext_panel1(admin->Panels,admin,cleanup);
                       break;
         case EBUSY:   fl_ext_panel1(admin->Panels+1,admin,cleanup);
                       break;
         case ENOBUFS: fl_ext_panel1(admin->Panels+2,admin,cleanup);
                       break;
         default:      fl_ext_panel1(admin->Panels+3,admin,cleanup);
                       break;
      }
      fl_do_forms();
   }
   admin->cleanup_code=2;
   
   if (((admin->ftstation_settings->mode=ioctl(admin->eib_fd,FT_GET_MODE))<0) ||
       ((admin->ftstation_settings->symb_baudrate=ioctl(admin->eib_fd,FT_GET_BAUDRATE))<0) ||
       ((len=ioctl(admin->eib_fd,FT_GET_PORT_LENGTH))<0) ||
       (ioctl(admin->eib_fd,FT_GET_OUTBUF_SIZE,&(admin->ftstation_settings->outbuf_size))<0) ||
       (ioctl(admin->eib_fd,FT_GET_WPBUF_SIZE,&(admin->ftstation_settings->wpbuf_size))<0) ||
       (ioctl(admin->eib_fd,FT_GET_RPBUF_SIZE,&(admin->ftstation_settings->rpbuf_size))<0) ||
       (ioctl(admin->eib_fd,FT_GET_RPBUF_MSGSIZE,&(admin->ftstation_settings->rpbuf_msgbuf_size))<0))
   {  switch(result=errno)
      {  case EBUSY:  fl_ext_panel1(admin->Panels+4,admin,cleanup);
                      break;
         case EFAULT: fl_ext_panel1(admin->Panels+5,admin,cleanup);
                      break;
         case EACCES: printf("no access rights\n");
                      break;
      }
      fl_do_forms();
   }
   
   admin->ftstation_settings->port=(char *)malloc((len+1)*sizeof(char));
   if ((ioctl(admin->eib_fd,FT_GET_PORT,admin->ftstation_settings->port)<0) ||
       (ioctl(admin->eib_fd,FT_RESET_STAT)<0))
   {  switch(result=errno)
      {  case EFAULT: fl_ext_panel1(admin->Panels+5,admin,cleanup);
                      break;
      }
      fl_do_forms();
   }
   time(&admin->stat_reset_time);
   admin->stat_get_time=admin->stat_reset_time;
   admin->act_interval=60;
   admin->stat_plan_time=admin->stat_get_time+admin->act_interval;   
      
   admin->MainForm=create_form_MainForm();
   admin->ResizeBufferForm=create_form_ResizeBufferForm();
   admin->MainForm->VisForm->u_vdata=admin;
   admin->ResizeBufferForm->VisForm->u_vdata=admin;
   
   fl_show_form(admin->MainForm->VisForm,FL_PLACE_CENTER,FL_TRANSIENT,"eibdrv_admin");
   admin->cleanup_code=3;
     
   fl_set_object_label(admin->MainForm->DeviceText,argv[1]);
   switch(admin->ftstation_settings->mode)
   {  case FT_STANDARDMODE: 
              fl_set_object_label(admin->MainForm->ModeText,"Standard Mode");
              break;
      case FT_SERVERMODE:
              fl_set_object_label(admin->MainForm->ModeText,"Server Mode");
              break;
   }
   
   i=0;
   while(i<=17)
   {  if (baudrates[i][0]==admin->ftstation_settings->symb_baudrate)
      {  sprintf(&BaudrateText[0],"%i",baudrates[i][1]);
         fl_set_object_label(admin->MainForm->BaudrateText,&BaudrateText[0]);
         break;
      }
      i++;
   }
   
   fl_set_object_label(admin->MainForm->PortText,admin->ftstation_settings->port);      

   fl_freeze_form(admin->MainForm->VisForm);   
   admin->ftstation_stat->overflow=1;
   actualize_MainForm_BufSiz(admin);
   actualize_MainForm_Stat(admin);
   actualize_MainForm_Status(admin,EIBDRV_ADMIN_STATUS_NOTSET);
   fl_unfreeze_form(admin->MainForm->VisForm);
   
   fl_set_timer(admin->MainForm->TimesTimer,1);
   fl_do_forms();
   
   return 0;
}



