/* ---------------------------------------------------------------------------
   eibdrv_sim.c
   ---------------------------------------------------------------------------
   eibdrv_sim Version 0.2.1
   Copyright (C) 2002, Wolfgang Tumfart
                       Donaustrasse 104/9
                       A-2344 Maria Enzersdorf
                       Austria (Europe)
                       tumfart@auto.tuwien.ac.at

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA.
   --------------------------------------------------------------------------- */

/*  ---------------------------------------------------------------------------
    This program serves as a benchmark program for the eibdrv driver
    environment to evaluate actual throughput and reliablity of the 
    eibdrv driver operating the FT1.2 protocol by simulating the driver
    load of multiple writing and reading user programs.
    
       * A certain number of write processes are forked, that write at random
         points in time within time intervals of certain length messages of
         random length within certain length limits to the eibdrv driver.
       * An equal number of read processes are forked, that collect messages
         from the eibdrv driver as soon as they are available.
         
    All processes are stopped and exited after a vertain run time and the 
    various data about throughput of the driver and the simulation system and 
    the theoretical upper limit are displayed in a panel and are written to
    a log file.
    
    All simulation parameter are chosen by the operator in the main window of 
    the program.
    --------------------------------------------------------------------------- */

/*  ---------------------------------------------------------------------------
    FormsLibrary is copyrighted by T.C. Zhao and Mark Overmars. It is not
    public domain, but may be used for running publically available free
    software.
    --------------------------------------------------------------------------- */

#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <termios.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/times.h>

#include "forms.h"
#include "eibdrv.h"
#include "eibdrv_fl.h"

#define EIBDRV_SIM_LOGFILE   "eibdrv_sim.log"     // File for result log
#define EIBDRV_SIM_PANELFILE "/sbin/eibdrv_sim.panel"   // File with panel messages
#define EIBDRV_SIM_PANELFILESIZE 12                 // Number of panels in panelfile

#define EIBDRV_SIM_NONBLOCKSTDWMODE 1          // Nonblocking standard write mode
#define EIBDRV_SIM_BLOCKSTDWMODE 2             // Blocking standard write mode
#define EIBDRV_SIM_NONBLOCKACKWMODE 3          // Nonblocking acknowledged write mode
#define EIBDRV_SIM_BLOCKACKWMODE 4             // Blocking acknowledged write mode

#define EIBDRV_SIM_INTERVAL 1                  // Fixed length time intervals
#define EIBDRV_SIM_REDINTERVAL 2               // Variable length time intervals

#define EIBDRV_SIM_W1_COL FL_PALEGREEN         // Bar for 1-Writes and 0-Reads
#define EIBDRV_SIM_W2_COL FL_DARKCYAN          // Bar for 2-Writes and 1-Reads
#define EIBDRV_SIM_W3_COL FL_ORCHID            // Bar for 3-Writes and 2-Reads
#define EIBDRV_SIM_WFAILED_COL FL_TOMATO       // Bar for failed-Writes

#define EIBDRV_SIM_STATE_INIT 1
#define EIBDRV_SIM_STATE_READY 2
#define EIBDRV_SIM_STATE_RUNNING 3
#define EIBDRV_SIM_STATE_SHUTDOWN 4

/*  Main Form */

typedef struct 
   { FL_FORM *VisForm;
     void *vdata;
     char *cdata;
     long  ldata;
     FL_OBJECT *DeviceText;
     FL_OBJECT *ModeText;
     FL_OBJECT *RunTimeCounter;
     FL_OBJECT *IntervalCounter;
     FL_OBJECT *MaxLengthCounter;
     FL_OBJECT *MinLengthCounter;
     FL_OBJECT *ProcessCounter;
     FL_OBJECT *NonblockingWModeCheckbtn;
     FL_OBJECT *BlockingWModeCheckbtn;
     FL_OBJECT *NonblockingAckWModeCheckbtn;
     FL_OBJECT *BlockingAckWModeCheckbtn;
     FL_OBJECT *WIntervalCheckbtn;
     FL_OBJECT  *WIntervalRedCheckbtn;
     FL_OBJECT *StartBtn;
     FL_OBJECT *ExitBtn;
   } FD_MainForm;

/* Form for display of simulation progress */

typedef struct 
   { FL_FORM *VisForm;
     void *vdata;
     char *cdata;
     long ldata;
     IndicatorType *Indicator;
     FL_OBJECT *StopBtn;
   } FD_RunForm;
   
   
/* Form for display of simulation results */

typedef struct
   { FL_FORM *VisForm;
   } FD_ResultsForm;

/*  Structure of shared memory segment used for communicating the
    simulation results between wirte/read processes and the split and
    main process.
 */
 
typedef struct
   { int result;
     unsigned long rw[EIB_BUF_ELEMENT_SIZE];          // Number of successful and 
     unsigned long w_failed[EIB_BUF_ELEMENT_SIZE];    // failed read and writes
     unsigned long tp;                                // Total penalty time in
   } Eib_bcu2_sim_shm;                                // reduced write interval mode

/*   All settings of the eibdrv driver that can be requested by
     a user program.
 */
 
typedef struct
   { int mode;                         // Standard/Server mode of driver
     int symb_baudrate;                // symbolic constant of baudrate
     char *port;                       // serial port 
     unsigned long outbuf_size;        // outgoing messages (writes)
     unsigned long wpbuf_size;         // writing processes
     unsigned long rpbuf_size;         // reading processes
     unsigned long rpbuf_msgbuf_size;  // incoming messages (reads)
   } Eib_bcu2_settings;

/*  All parameter of the simulation current configured in the
    main form.
 */
 
typedef struct 
   { int run_time;           // Run time of all read and write processes
     int write_interval;     // Time interval within msg are written randomly
     int min_len;            // Minimal value for random message length
     int max_len;            // Maximal value for random message length
     int proc_count;         // Number of write/read processes to be forked
     int wmode;              // Blocking, nonblocking or acknowledged write
     int winterval;          // W or w/o penalty reduction of write time interval
   } Eib_bcu2_sim_parameter;
   
/*  Structure of the simulator */
   
typedef struct 
   { FD_MainForm *MainForm;       // Main form
     FD_RunForm *RunForm;         // Progress visualization form
     PanelContent *Panels;        // Message panels
     char *device;                // Device string
     int f_stop_readwrite;        // Flag for run timeout of write/read processes
     int state;                   // Internal state of simulator
     time_t start_time;           // Simulation start time
     time_t last_act_time;        // Time of last indicator actualization
     int eib_fd;                  // Filedescriptor for driver 
     FILE *log_fd;                // Filedescriptor for logfile
     Eib_bcu2_settings *ftstation_settings; // Driver settings
     Eib_bcu2_sim_parameter *act_parameter; // Parameter set in main form
     Eib_bcu2_sim_parameter *sim_parameter; // Parameter for simulation
     Eib_statistics *stat;        // Driver statistics after simulation run
     time_t stat_reset_time;      // Time since last statistics reset
     time_t stat_get_time;        // Time since last statistics actualization
     pid_t split_pid;             // Process-ID of read/write-generating process
     int cleanup_code;            // Abort of program
     
     int shm_results_id;               // ID of shared memory for results
     Eib_bcu2_sim_shm *shm_results;    // Shared memory for results
     int sem_shm_results;              // sem synchronizing access to result shm
     int sem_start_rw;                 // sem synchronizing start of read/writes
   } Eib_bcu2_sim;

/* Structure of read/write processes */

typedef struct
   { int eib_fd;                       // Filedescriptor for driver
     Eib_bcu2_sim_shm *shm_segment;    // Segment in shm for process results
     Eib_bcu2_sim_shm local_results;   // Local results
   } Eib_bcu2_sim_rw;             
   
     
Eib_bcu2_sim *sim;      // Central simulator struct used by all processes
Eib_bcu2_sim_rw rw_sim; // Read/write process struct 

int baudrates[18][2]={ {B50,50},       // Table of symbolic baudrates and their integer 
                       {B75,75},       // value
                       {B110,110},
                       {B134,134},
                       {B150,150},
                       {B200,200},
                       {B300,300},
                       {B600,600},
                       {B1200,1200},
                       {B1800,1800},
                       {B2400,2400},
                       {B4800,4800}, 
                       {B9600,9600},
                       {B19200,19200},
                       {B38400,38400},
                       {B57600,57600},
                       {B115200,115200},
                       {B230400,230400}};


/*  --------------------------------------------------------------------------- 
    Various function prototypes.
    ---------------------------------------------------------------------------
*/
void cleanup(void *arg,int btn);
void panel_cb(void *arg,int btn);
void sim_split();
void main_sigchld_handler();
void StopBtn_cb(FL_OBJECT *obj,long data);

/*  --------------------------------------------------------------------------- 
    Allocation and deallocation of simulator data structures,
    ---------------------------------------------------------------------------
 */

Eib_bcu2_sim *getEib_bcu2_sim()
{  Eib_bcu2_sim *help=(Eib_bcu2_sim *)fl_calloc(1,sizeof(Eib_bcu2_sim));

   help->MainForm=NULL;
   help->RunForm=NULL;
   help->Panels=NULL;
   help->state=EIBDRV_SIM_STATE_INIT;
   help->ftstation_settings=(Eib_bcu2_settings *)fl_calloc(1,sizeof(Eib_bcu2_settings));
   help->ftstation_settings->port=NULL;
   help->act_parameter=(Eib_bcu2_sim_parameter *)fl_calloc(1,sizeof(Eib_bcu2_sim_parameter));
   help->sim_parameter=(Eib_bcu2_sim_parameter *)fl_calloc(1,sizeof(Eib_bcu2_sim_parameter));
   help->stat=(Eib_statistics *)fl_calloc(1,sizeof(Eib_statistics));
   help->cleanup_code=0;

   return help;
}

void returnEib_bcu2_sim(Eib_bcu2_sim *sim)
{  free(sim->MainForm);
   free(sim->RunForm);
   returnPanelContentA(sim->Panels,EIBDRV_SIM_PANELFILESIZE);
   if (sim->ftstation_settings->port!=NULL) free(sim->ftstation_settings->port);
   free(sim->ftstation_settings);
   free(sim->act_parameter);
   free(sim->sim_parameter);
   free(sim->stat);
   free(sim->device);
   free(sim);
}

/*  This function delays the caller process for usec Mikroseconds 
    by blocking it.
 */
 
void delay(unsigned long usec)
{  struct timeval timeout;
   
   if (usec==0) return;
   timeout.tv_sec=usec / 1000000L;
   timeout.tv_usec=usec % 1000000L;
   
   select(0,NULL,NULL,NULL,&timeout);
}


/*  This function modifies the Text1 variable of the given panel
    structure to display a given error_code in the message text
    of a message panel.
 */
 
void make_InternalErrorPanel(PanelContent *panel,int error_code)
{  char helpstring[40];
   
   if (panel==NULL) return;
   
   sprintf(&helpstring[0],"Internal error %i occured!",error_code);
   free(panel->Text1);
   panel->Text1=(char *)malloc((strlen(&helpstring[0])+1)*sizeof(char));
   strcpy(panel->Text1,&helpstring[0]);
}

   
/*  --------------------------------------------------------------------------- 
    Creation of all forms and callbacks of all visual elements.
    ---------------------------------------------------------------------------
 */
   
char *get_wmode_text(int wmode)
{  switch(wmode)
   {  case EIBDRV_SIM_NONBLOCKSTDWMODE: return "Nonblocking Standard";
      case EIBDRV_SIM_BLOCKSTDWMODE:    return "Blocking Standard";
      case EIBDRV_SIM_NONBLOCKACKWMODE: return "Nonblocking Acknowledged";
      case EIBDRV_SIM_BLOCKACKWMODE:    return "Blocking Acknowledged";
   }
   return NULL;
}

char *get_winterval_text(int winterval)
{  switch(winterval)
   {  case EIBDRV_SIM_INTERVAL:    return "Without penalty reduction";
      case EIBDRV_SIM_REDINTERVAL: return "With penalty reduction";
   }
   return NULL;
}

void panel_cb(void *arg,int btn)
{}

void Null_cb(FL_OBJECT *obj,long data)
{}

void create_object_copyright(int x,int y)
{  FL_OBJECT *obj;
   
   fl_ext_formobjects_add_box(FL_EMBOSSED_BOX,x,y,135,80,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,x+5,y+5,125,30,"eibdrv_sim",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lcolor(obj,FL_BLUE);
      fl_set_object_lsize(obj,FL_LARGE_SIZE);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_EMBOSSED_STYLE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,x+20,y+30,110,20,"Version 0.2.1",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lsize(obj,FL_NORMAL_SIZE);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_BOLD_STYLE);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,x+10,y+50,120,20,"(C) Wolfgang Tumfart",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
}



void CloseBtn_cb(FL_OBJECT *obj,long data)
{  fl_hide_form(obj->form);
   fl_free_form(obj->form);
}
   

FD_ResultsForm *create_form_ResultsForm(Eib_bcu2_sim *sim)
{  FL_OBJECT *obj;
   FD_ResultsForm *fdui = (FD_ResultsForm *)fl_calloc(1,sizeof(*fdui));
   char helpstring[100];
   int i,j,baudrate_index;
   unsigned long w_upper_limit;
   unsigned long w_proc_success=0;
   unsigned long w_proc_total=0;
   unsigned long w_proc_failed=0;
   float w_proc_total_percent=0;
   float w_proc_success_percent=0;
   unsigned long w_drv_total=0;
   unsigned long w_drv_success=0;
   float w_drv_total_percent=0;
   float w_drv_success_percent=0;
   unsigned long w_max=0;
   unsigned long r_proc_total=0;
   unsigned long r_drv_total=0;
   unsigned long r_max=0;
   double h;
   int h_failed,h_rw;
   Eib_bcu2_sim_shm local_w_results;
   Eib_bcu2_sim_shm local_r_results;

   double diff_time;
   unsigned long rw_ges;
   double rw_pt;
   int lowerbound;
   int avg_w1time;

   /* Summarizing the write and read results over all processes */
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  local_w_results.rw[i]=0;
      local_w_results.w_failed[i]=0;
      for (j=0;j<sim->sim_parameter->proc_count;j++)
      {  local_w_results.rw[i]+=(sim->shm_results+j)->rw[i];
         local_w_results.w_failed[i]+=(sim->shm_results+j)->w_failed[i];
      }
   }
   
   local_w_results.tp=0;
   for (j=0;j<sim->sim_parameter->proc_count;j++)
      local_w_results.tp+=(sim->shm_results+j)->tp;

   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  local_r_results.rw[i]=0;
      for (j=0;j<sim->sim_parameter->proc_count;j++)
         local_r_results.rw[i]+=(sim->shm_results+sim->sim_parameter->proc_count+j)->rw[i];
   }
         
   baudrate_index=0;
   while(baudrate_index<=17)
   {  if (baudrates[baudrate_index][0]==sim->ftstation_settings->symb_baudrate)
         break;
      baudrate_index++;
   }

   /* Calculation of throughput */
   w_upper_limit=(unsigned long)(((double)sim->sim_parameter->run_time*60000)/
                 sim->sim_parameter->write_interval)*sim->sim_parameter->proc_count;
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  w_proc_success+=local_w_results.rw[i];
      w_proc_failed+=local_w_results.w_failed[i];
   }
   w_proc_total+=(w_proc_failed+w_proc_success);
   
   for (i=1;i<4;i++)
      w_drv_success+=sim->stat->write[i];
   w_drv_total=w_drv_success+sim->stat->write[0];

   if (w_upper_limit>0) 
   {  w_proc_total_percent=((float)w_proc_total)/w_upper_limit*100;
      w_drv_total_percent=((float)w_drv_total)/w_upper_limit*100;
   }
   if (w_proc_total>0) w_proc_success_percent=((float)w_proc_success)/w_proc_total*100;
   if (w_drv_total>0)  w_drv_success_percent=((float)w_drv_success)/w_drv_total*100;

   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
      r_proc_total+=local_r_results.rw[i];
   
   for (i=0;i<3;i++)
      r_drv_total+=sim->stat->read[i];

   fdui->VisForm=fl_bgn_form(FL_NO_BOX,980,675);
   fl_ext_formobjects_add_box(FL_UP_BOX,0,0,980,675,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,20,405,115,"Settings",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,30,55,20,"Device",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,90,30,315,20,sim->device,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,55,55,20,"Mode",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,90,55,315,20,
         (sim->ftstation_settings->mode==FT_STANDARDMODE ? "Standard Mode" : "Server Mode"),
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,80,55,20,"Baudrate",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",baudrates[baudrate_index][1]);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,90,80,315,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,105,55,20,"Port",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,90,105,315,20,sim->ftstation_settings->port,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
  
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,155,190,65,"Process Buffer Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,165,45,20,"WRITE",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,190,45,20,"READ",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,165,45,20,"Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,80,190,45,20,"Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",sim->ftstation_settings->wpbuf_size);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,115,165,40,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",sim->ftstation_settings->rpbuf_size);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,115,190,40,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,235,155,190,65,"Message Buffer Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,245,165,45,20,"WRITE",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,245,190,45,20,"READ",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,295,165,45,20,"Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,295,190,45,20,"Size",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",sim->ftstation_settings->outbuf_size);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,330,165,40,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",sim->ftstation_settings->rpbuf_msgbuf_size);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,330,190,40,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,445,20,355,200,"Simulation Parameter",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,30,155,20,"Run time (min):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",sim->sim_parameter->run_time);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,30,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,55,155,20,"Write interval (ms):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",sim->sim_parameter->write_interval);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,55,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,80,155,20,"Minimal message length (byte):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",sim->sim_parameter->min_len);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,80,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,105,155,20,"Maximal message length (byte):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",sim->sim_parameter->max_len);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,105,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,130,155,20,"Number of processes:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%i",sim->sim_parameter->proc_count);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,130,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,155,155,20,"Write Mode:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,155,165,20,get_wmode_text(sim->sim_parameter->wmode),
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,180,155,20,"Write Interval:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,615,180,165,20,get_winterval_text(sim->sim_parameter->winterval),
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   create_object_copyright(820,20);      

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,240,405,265,"Write (Process Simulation)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,250,105,20,"Upper Limit:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",w_upper_limit);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,140,250,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,275,105,20,"Process Throughput:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",w_proc_total);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,140,275,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%3.2f %%",w_proc_total_percent);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,220,275,55,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",w_proc_success);         
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,280,275,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%3.2f %%",w_proc_success_percent);         
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,360,275,55,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,300,105,20,"Driver Throughput:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",w_drv_total);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,140,300,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%3.2f %%",w_drv_total_percent);         
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,220,300,55,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",w_drv_success);         
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,280,300,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%3.2f %%",w_drv_success_percent);         
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,360,300,55,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,30,330,115,20,"Total penalty time (ms):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",local_w_results.tp);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,170,330,90,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         
   fl_ext_formobjects_add_box(FL_BORDER_BOX,50,360,345,115,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  sprintf(&helpstring[0],"%i",i+1);
      fl_ext_formobjects_add_text(&obj,FL_NO_BOX,50+15*i,475,15,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         fl_set_object_lsize(obj,FL_TINY_SIZE);
         fl_set_object_lalign(obj,FL_ALIGN_CENTER);
      if (local_w_results.rw[i]+local_w_results.w_failed[i]>w_max)
         w_max=local_w_results.rw[i]+local_w_results.w_failed[i];
   }
   h=115.0/w_max;
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  h_failed=(int)((double)local_w_results.w_failed[i]*h);
      if (h_failed>0)
      {  fl_ext_formobjects_add_box2(&obj,FL_FRAME_BOX,50+15*i,475-h_failed,15,h_failed,"",
               FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
            fl_set_object_color(obj,FL_TOMATO,BOX_COL2);
      }
      h_rw=(int)((double)local_w_results.rw[i]*h);
      if (h_rw>0)
      {  fl_ext_formobjects_add_box2(&obj,FL_FRAME_BOX,50+15*i,475-h_failed-h_rw,15,h_rw,"",
               FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
            fl_set_object_color(obj,FL_WHEAT,BOX_COL2);
      }
   }      

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,445,240,405,265,"Read (Process Simulation)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,275,105,20,"Process Throughput:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",r_proc_total);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,565,275,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,455,300,105,20,"Driver Throughput:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li",r_drv_total);
   fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,565,300,75,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         
   fl_ext_formobjects_add_box(FL_BORDER_BOX,475,360,345,115,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  sprintf(&helpstring[0],"%i",i+1);
      fl_ext_formobjects_add_text(&obj,FL_NO_BOX,475+15*i,475,15,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         fl_set_object_lsize(obj,FL_TINY_SIZE);
         fl_set_object_lalign(obj,FL_ALIGN_CENTER);
      if (local_r_results.rw[i]>r_max)
         r_max=local_r_results.rw[i];
   }
   h=115.0/r_max;
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
   {  h_rw=(int)((double)local_r_results.rw[i]*h);
      if (h_rw>0)
      {  fl_ext_formobjects_add_box2(&obj,FL_FRAME_BOX,475+15*i,475-h_rw,15,h_rw,"",
               FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
            fl_set_object_color(obj,FL_WHEAT,BOX_COL2);
      }
   }      

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,525,405,130,"Write (Driver Statistics)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,535,165,20,"1 - 2 - 3 attempts - failed",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,345,575,60,20,"total writes",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,345,590,60,20,"writes / sec",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,445,525,405,130,"Read (Driver Statistics)",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);            
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,535,130,20,"0 - 1 - 2 retransmissions",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,770,575,60,20,"total reads",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,770,590,60,20,"reads / sec",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);

   fl_ext_formobjects_add_box(FL_BORDER_BOX,40,555,365,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   rw_ges=0;
   for (i=0;i<4;i++) rw_ges+=sim->stat->write[i];
   rw_pt=363.0/rw_ges;
   lowerbound=41;
   for (i=1;i<5;i++)
   {  if (sim->stat->write[i%4]>0)
      {  fl_ext_formobjects_add_box2(&obj,FL_FLAT_BOX,lowerbound,556,404-lowerbound,18,"",
               FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         switch(i%4)
         {  case 1: fl_set_object_color(obj,EIBDRV_SIM_W1_COL,BOX_COL2); break;
            case 2: fl_set_object_color(obj,EIBDRV_SIM_W2_COL,BOX_COL2); break;
            case 3: fl_set_object_color(obj,EIBDRV_SIM_W3_COL,BOX_COL2); break;
            case 0: fl_set_object_color(obj,EIBDRV_SIM_WFAILED_COL,BOX_COL2); break;
         }
         lowerbound+=(int)(sim->stat->write[i%4]*rw_pt);
      }
   }

   fl_ext_formobjects_add_box(FL_BORDER_BOX,465,555,365,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);   
   rw_ges=0;
   for (i=0;i<3;i++) rw_ges+=sim->stat->read[i];
   rw_pt=363.0/rw_ges;
   lowerbound=466;
   for (i=0;i<3;i++)
   {  if (sim->stat->read[i]>0)
      {  fl_ext_formobjects_add_box2(&obj,FL_FLAT_BOX,lowerbound,556,829-lowerbound,18,"",
               FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
         switch(i)
         {  case 0: fl_set_object_color(obj,EIBDRV_SIM_W1_COL,BOX_COL2); break;
            case 1: fl_set_object_color(obj,EIBDRV_SIM_W2_COL,BOX_COL2); break;
            case 2: fl_set_object_color(obj,EIBDRV_SIM_W3_COL,BOX_COL2); break;
         }
         lowerbound+=(int)(sim->stat->read[i]*rw_pt);
      }
   }

   diff_time=difftime(sim->stat_get_time,sim->stat_reset_time);
   sprintf(&helpstring[0],"%li - %li - %li - %li",sim->stat->write[1],sim->stat->write[2],
         sim->stat->write[3],sim->stat->write[0]);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,575,165,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   sprintf(&helpstring[0],"%li - %li - %li",sim->stat->read[0],sim->stat->read[1],sim->stat->read[2]);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,575,205,20,&helpstring[0],
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);

   if (diff_time>0)
   {  sprintf(&helpstring[0],"%3.2f - %3.2f - %3.2f - %3.2f",sim->stat->write[1]/diff_time,
            sim->stat->write[2]/diff_time,sim->stat->write[3]/diff_time,sim->stat->write[0]/diff_time);
      fl_ext_formobjects_add_text(&obj,FL_NO_BOX,40,590,205,20,&helpstring[0],
            FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      sprintf(&helpstring[0],"%3.2f - %3.2f - %3.2f",sim->stat->read[0]/diff_time,
            sim->stat->read[1]/diff_time,sim->stat->read[2]/diff_time);
      fl_ext_formobjects_add_text(&obj,FL_NO_BOX,465,590,205,20,&helpstring[0],
            FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   }

   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,615,140,15,"Average total time for write",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,630,170,15,"with success in first attempt (ms):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   if (sim->stat->write[1]>0)   
   {  avg_w1time=((sim->stat->total_w1time.tv_sec+sim->stat->total_w1time.tv_usec/1000000.0)*1000)/
             sim->stat->write[1];
      if (avg_w1time<1000)
         sprintf(&helpstring[0],"%i",avg_w1time);
      else
         strcpy(&helpstring[0],"");
      fl_ext_formobjects_add_text(&obj,FL_BORDER_BOX,205,625,75,20,&helpstring[0],
            FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   }      

   fl_ext_formobjects_add_btn(&obj,FL_NORMAL_BUTTON,870,630,90,25,"Close",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(obj,CloseBtn_cb,0);
      
   fl_end_form();
   
   fdui->VisForm->fdui=fdui;
   return fdui;   
}


#if FL_INCLUDE_VERSION >= 89
void idle_Indicator(int arg1,void *arg2)
#else 
int idle_Indicator(XEvent *xev,void *userdata)
#endif
{  time_t actual_time;
   time_t stop_time=sim->start_time+sim->sim_parameter->run_time*60;
   
   time(&actual_time);
   if (actual_time!=sim->last_act_time)
   {  if (actual_time>=stop_time)
         fl_ext_update_indicator(sim->RunForm->Indicator,FL_EXT_INDICATOR_RUNNING,100);
      else
         fl_ext_update_indicator(sim->RunForm->Indicator,FL_EXT_INDICATOR_RUNNING,
            (100.0*difftime(actual_time,sim->start_time))/(sim->sim_parameter->run_time*60));
      sim->last_act_time=actual_time;
   }

#if FL_INCLUDE_VERSION >= 89         
   fl_add_timeout(500,idle_Indicator,NULL);
#else
   return 0;
#endif
}


FD_RunForm *create_form_RunForm(Eib_bcu2_sim *sim)
{  FL_OBJECT *obj;
   int form_width=FL_EXT_INDICATOR_MINWIDTH+160;
   int form_height=FL_EXT_INDICATOR_MINHEIGHT+40;
   FD_RunForm *fdui=(FD_RunForm *)fl_calloc(1,sizeof(*fdui));
   
   fdui->VisForm=fl_bgn_form(FL_NO_BOX,form_width,form_height);
   fl_ext_formobjects_add_box(FL_UP_BOX,0,0,form_width,form_height,"",
      FL_RESIZE_NONE,FL_NorthWest,FL_SouthEast);
   fl_ext_formobjects_add_btn(&obj,FL_NORMAL_BUTTON,form_width-110,form_height-45,90,25,"Stop",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(obj,StopBtn_cb,0);
   fl_end_form();

   fdui->VisForm->fdui=fdui;
   fdui->VisForm->u_vdata=sim;
      
   fdui->Indicator=fl_ext_add_indicator(fdui->VisForm,20,20,
         FL_EXT_INDICATOR_MINWIDTH,FL_EXT_INDICATOR_MINHEIGHT,FL_NoGravity);
   
   return fdui;   
}


void RunTimeCounter_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   sim->act_parameter->run_time=fl_get_counter_value(obj);
}


void IntervalCounter_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   sim->act_parameter->write_interval=fl_get_counter_value(obj);
}


void MinLengthCounter_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   sim->act_parameter->min_len=fl_get_counter_value(obj);
   fl_set_counter_bounds(sim->MainForm->MaxLengthCounter,fl_get_counter_value(obj),23);
}


void MaxLengthCounter_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   sim->act_parameter->max_len=fl_get_counter_value(obj);
   fl_set_counter_bounds(sim->MainForm->MinLengthCounter,1,fl_get_counter_value(obj));
}


void ProcessCounter_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   sim->act_parameter->proc_count=fl_get_counter_value(obj);
}


void WModeCheckbtn_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;
   
   sim->act_parameter->wmode=(int)data;
}


void WIntervalCheckbtn_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;
   
   sim->act_parameter->winterval=(int)data;
}
      

void StartBtn_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;
   int result;
      
   fl_ext_deactivate_object(sim->MainForm->StartBtn);

   *(sim->sim_parameter)=*(sim->act_parameter);

   sim->cleanup_code=6;
   if ((sim->shm_results_id=shmget(IPC_PRIVATE,2*sim->sim_parameter->proc_count*sizeof(Eib_bcu2_sim_shm),
              0666|IPC_CREAT|IPC_EXCL))==-1)
   {  make_InternalErrorPanel(sim->Panels+6,56);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      return;
   }
   sim->cleanup_code=7;

   if ((sim->shm_results=(Eib_bcu2_sim_shm *)shmat(sim->shm_results_id,NULL,0))==(void *)-1)
   {  make_InternalErrorPanel(sim->Panels+6,57);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      return;
   }
   sim->cleanup_code=8;

   if (ioctl(sim->eib_fd,FT_RESET_STAT)<0)
   {  switch(result=errno)
      {  case EBUSY:  fl_ext_panel1(sim->Panels+4,sim,cleanup);
                      break;
         case EACCES: fl_ext_panel1(sim->Panels+10,sim,cleanup);
                      break;
      }
      return;
   }
   time(&sim->stat_reset_time);

   switch(sim->split_pid=fork())
   {  case  0: sim_split();
      case -1: make_InternalErrorPanel(sim->Panels+6,58);
               fl_ext_panel1(sim->Panels+6,sim,cleanup);
               return;
      default: sim->state=EIBDRV_SIM_STATE_RUNNING;
               sim->cleanup_code=5;
               break;
   }
   
   fl_ext_reset_indicator(sim->RunForm->Indicator);
   fl_ext_start_indicator(sim->RunForm->Indicator);
   time(&sim->start_time);
   fl_ext_update_indicator(sim->RunForm->Indicator,FL_EXT_INDICATOR_RUNNING,0);
   fl_show_form(sim->RunForm->VisForm,FL_PLACE_CENTER,FL_TRANSIENT,"eibdrv_sim");

#if FL_INCLUDE_VERSION >= 89
   fl_add_timeout(500,idle_Indicator,NULL);
#else
   fl_set_idle_callback(idle_Indicator,0);
#endif
   
}


void StopBtn_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;
   if (sim->state==EIBDRV_SIM_STATE_RUNNING)
      kill(-sim->split_pid,SIGTERM);
}


void ExitBtn_cb(FL_OBJECT *obj,long data)
{  Eib_bcu2_sim *sim=(Eib_bcu2_sim *)obj->form->u_vdata;

   fl_hide_form(obj->form);
   if (sim->state==EIBDRV_SIM_STATE_RUNNING)
   {  fl_hide_form(sim->RunForm->VisForm);
#if FL_INCLUDE_VERSION < 89      
      fl_set_idle_callback(0,0);
#endif
      sim->state=EIBDRV_SIM_STATE_SHUTDOWN;    // wait for all processes in
      kill(-sim->split_pid,SIGTERM);             // main_sigchld and exit
   }
   else
   {  semctl(sim->sem_start_rw,0,IPC_RMID,(int)0);    // direct exit because no
      semctl(sim->sem_shm_results,0,IPC_RMID,(int)0); // read or write processes 
      fl_free_form(sim->MainForm->VisForm);           // are running
      fl_ext_free_indicator(sim->RunForm->Indicator);
      fl_free_form(sim->RunForm->VisForm);
      close(sim->eib_fd);
      returnEib_bcu2_sim(sim);
      fl_finish();
      exit(0);
   }
}


FD_MainForm *create_form_MainForm(Eib_bcu2_sim *sim)
{
   FL_OBJECT *obj;
   FD_MainForm *fdui = (FD_MainForm *) fl_calloc(1, sizeof(*fdui));

   fdui->VisForm = fl_bgn_form(FL_NO_BOX, 520, 425);
   fl_ext_formobjects_add_box(FL_UP_BOX,0,0,520,425,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_SouthEast);
  
   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,20,325,80,"Settings",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,30,40,20,"Device:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,55,40,20,"Mode:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&fdui->DeviceText,FL_BORDER_BOX,80,30,250,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&fdui->ModeText,FL_BORDER_BOX,80,55,250,20,"",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);

   create_object_copyright(365,20);      

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,120,370,145,"Simulation Parameter",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,130,160,20,"Run Time (min):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_counter(&fdui->RunTimeCounter,FL_NORMAL_COUNTER,200,130,170,20,"",
         FL_ALIGN_LEFT,1,1000,1,10,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_counter_value(fdui->RunTimeCounter, 1);
      sim->act_parameter->run_time=1;
      fl_set_object_callback(fdui->RunTimeCounter,RunTimeCounter_cb,0);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,155,160,20,"Write Interval (ms):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_counter(&fdui->IntervalCounter,FL_NORMAL_COUNTER,200,155,170,20,"",
         FL_ALIGN_LEFT,10,10000,1,100,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_counter_value(fdui->IntervalCounter, 1000);
      sim->act_parameter->write_interval=1000;
      fl_set_object_callback(fdui->IntervalCounter,IntervalCounter_cb,0);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,180,160,20,"Minimal message length (Byte):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_counter(&fdui->MinLengthCounter,FL_SIMPLE_COUNTER,200,180,170,20,"",
         FL_ALIGN_LEFT,1,23,1,1,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_counter_value(fdui->MinLengthCounter, 1);
      sim->act_parameter->min_len=1;
      fl_set_object_callback(fdui->MinLengthCounter,MinLengthCounter_cb,0);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,205,160,20,"Maximal message length (Byte):",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_counter(&fdui->MaxLengthCounter,FL_SIMPLE_COUNTER,200,205,170,20,"",
         FL_ALIGN_LEFT,1,EIB_BUF_ELEMENT_SIZE,1,1,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_counter_value(fdui->MaxLengthCounter,EIB_BUF_ELEMENT_SIZE);
      sim->act_parameter->max_len=EIB_BUF_ELEMENT_SIZE;
      fl_set_object_callback(fdui->MaxLengthCounter,MaxLengthCounter_cb,0);
   fl_ext_formobjects_add_text(&obj,FL_NO_BOX,35,230,160,20,"Number of processes:",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_ext_formobjects_add_counter(&fdui->ProcessCounter,FL_SIMPLE_COUNTER,200,230,170,20,"",
         FL_ALIGN_LEFT,1,100,1,10,0,
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_counter_value(fdui->ProcessCounter, 1);
      sim->act_parameter->proc_count=1;
      fl_set_object_callback(fdui->ProcessCounter,ProcessCounter_cb,0);

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,20,290,180,115,"Write Mode",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_bgn_group();
   fl_ext_formobjects_add_checkbtn(&fdui->NonblockingWModeCheckbtn,FL_RADIO_BUTTON,
         30,300,130,20,"Nonblocking Standard",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->NonblockingWModeCheckbtn,WModeCheckbtn_cb,EIBDRV_SIM_NONBLOCKSTDWMODE);
   fl_ext_formobjects_add_checkbtn(&fdui->BlockingWModeCheckbtn,FL_RADIO_BUTTON,
         30,325,130,20,"Blocking Standard",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->BlockingWModeCheckbtn,WModeCheckbtn_cb,EIBDRV_SIM_BLOCKSTDWMODE);
   fl_ext_formobjects_add_checkbtn(&fdui->NonblockingAckWModeCheckbtn,FL_RADIO_BUTTON,
         30,350,130,20,"Nonblocking Acknowledged",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->NonblockingAckWModeCheckbtn,WModeCheckbtn_cb,EIBDRV_SIM_NONBLOCKACKWMODE);
   fl_ext_formobjects_add_checkbtn(&fdui->BlockingAckWModeCheckbtn,FL_RADIO_BUTTON,
         30,375,130,20,"Blocking Acknowledged",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->BlockingAckWModeCheckbtn,WModeCheckbtn_cb,EIBDRV_SIM_BLOCKACKWMODE);
      fl_set_button(fdui->NonblockingAckWModeCheckbtn,1);
      sim->act_parameter->wmode=EIBDRV_SIM_NONBLOCKACKWMODE;
   fl_end_group();

   fl_ext_formobjects_add_labelframe(FL_ENGRAVED_FRAME,220,290,170,115,"Write Interval",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
   fl_bgn_group();
   fl_ext_formobjects_add_checkbtn(&fdui->WIntervalCheckbtn,FL_RADIO_BUTTON,
         230,300,130,20,"Without Penalty Reduction",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->WIntervalCheckbtn,WIntervalCheckbtn_cb,EIBDRV_SIM_INTERVAL);
      fl_set_button(fdui->WIntervalCheckbtn,1);      
      sim->act_parameter->winterval=EIBDRV_SIM_INTERVAL;
   fl_ext_formobjects_add_checkbtn(&fdui->WIntervalRedCheckbtn,FL_RADIO_BUTTON,
         230,325,130,20,"With Penalty Reduction",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->WIntervalRedCheckbtn,WIntervalCheckbtn_cb,EIBDRV_SIM_REDINTERVAL);
   fl_end_group();

   fl_ext_formobjects_add_btn(&fdui->StartBtn,FL_NORMAL_BUTTON,
         410,345,90,25,"Start",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->StartBtn,StartBtn_cb,0);
   fl_ext_formobjects_add_btn(&fdui->ExitBtn,FL_NORMAL_BUTTON,
         410,380,90,25,"Exit",
         FL_RESIZE_NONE,FL_NorthWest,FL_NorthWest);
      fl_set_object_callback(fdui->ExitBtn,ExitBtn_cb,0);

   fl_end_form();
   fdui->VisForm->fdui = fdui;
   fdui->VisForm->u_vdata=sim;
   return fdui;
}


int create_Log(Eib_bcu2_sim *sim)
{
   /*  This function write both driver statistics and simulation results
       to a logfile. It returs -1, if an error occured or 0 otherwise.
    */ 

   int i,j;
   time_t log_time;
   char line[96];
   char singleline[218];
   int avg_w1time;
          
   if ((sim->log_fd=fopen(EIBDRV_SIM_LOGFILE,"w"))==NULL)
      return -1;

   time(&log_time);
   for (i=0;i<95;i++) line[i]='='; line[95]='\0';
   for (i=0;i<218;i++) singleline[i]='-'; singleline[218]='\0';

   i=0;
   while(i<=17)
   {  if (baudrates[i][0]==sim->ftstation_settings->symb_baudrate)
         break;
      i++;
   }

   fprintf(sim->log_fd,"%s\nLOG for: %s\nFile:    %s\nDate:    %s%s\n",
         line,"eibdrv_sim","eibdrv_sim.log",ctime(&log_time),line);
   fprintf(sim->log_fd,"\n\nCONFIGURATION:\n\nDevice:                        %s\n",sim->device);
   fprintf(sim->log_fd,"Mode:                          %s\n",
         (sim->ftstation_settings->mode==FT_STANDARDMODE ? "Standard Mode" : "Server Mode"));
   fprintf(sim->log_fd,"Baudrate:                      %i\n",baudrates[i][1]);
   fprintf(sim->log_fd,"Port:                          %s\n",sim->ftstation_settings->port);

   fprintf(sim->log_fd,"Process buffer size - Write:   %li\n",sim->ftstation_settings->wpbuf_size);
   fprintf(sim->log_fd,"                    - Read:    %li\n",sim->ftstation_settings->rpbuf_size);
   fprintf(sim->log_fd,"Message buffer size - Write:   %li\n",sim->ftstation_settings->outbuf_size);
   fprintf(sim->log_fd,"                    - Read:    %li\n%s",
         sim->ftstation_settings->rpbuf_msgbuf_size,line);
   fprintf(sim->log_fd,"\n\nSIMULATION PARAMETER:\n\n");
   fprintf(sim->log_fd,"Run time (min):                %i\n",sim->sim_parameter->run_time);
   fprintf(sim->log_fd,"Write interval (ms):           %i\n",sim->sim_parameter->write_interval);
   fprintf(sim->log_fd,"Minimal message length (byte): %i\n",sim->sim_parameter->min_len);
   fprintf(sim->log_fd,"Maximal message length (byte): %i\n",sim->sim_parameter->max_len);
   fprintf(sim->log_fd,"Number of processes:           %i\n",sim->sim_parameter->proc_count);
   fprintf(sim->log_fd,"Write mode:                    %s\n",get_wmode_text(sim->sim_parameter->wmode));
   fprintf(sim->log_fd,"Write interval:                %s\n",get_winterval_text(sim->sim_parameter->winterval));
   fprintf(sim->log_fd,"%s\n\n",line);

   fprintf(sim->log_fd,"%s\nWrite        Result  Rest          Message Length (Byte)\n",singleline);
   fprintf(sim->log_fd,"Process No.          Penalty Time  ");
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++) fprintf(sim->log_fd,"%7i ",i+1);
   fprintf(sim->log_fd,"\n%s",singleline);
   for (j=0;j<sim->sim_parameter->proc_count;j++)
   {  fprintf(sim->log_fd,"\n   %4i      success   %9li   ",j+1,(sim->shm_results+j)->tp);
      for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
         fprintf(sim->log_fd,"%7li ",(sim->shm_results+j)->rw[i]);
      fprintf(sim->log_fd,"\n             failed                ");
      for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
         fprintf(sim->log_fd,"%7li ",(sim->shm_results+j)->w_failed[i]);
      fprintf(sim->log_fd,"\n");
   }        
   fprintf(sim->log_fd,"%s\n\n",singleline);

   fprintf(sim->log_fd,"%s\nRead                               Message Length (Byte)\n",singleline);
   fprintf(sim->log_fd,"Process No.                        ");
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++) fprintf(sim->log_fd,"%7i ",i+1);
   fprintf(sim->log_fd,"\n%s",singleline);
   for (j=0;j<sim->sim_parameter->proc_count;j++)
   {  fprintf(sim->log_fd,"\n   %4i                            ",j+1);
      for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++)
         fprintf(sim->log_fd,"%7li ",(sim->shm_results+sim->sim_parameter->proc_count+j)->rw[i]);
   }        
   fprintf(sim->log_fd,"\n%s\n",singleline);
   
   fprintf(sim->log_fd,"\n\nWrite Messages (1/2/3 attemps/failed):  %9li %9li %9li %9li\n",
         sim->stat->write[1],sim->stat->write[2],sim->stat->write[3],sim->stat->write[0]);
   fprintf(sim->log_fd,"Read Messages  (0/1/2 retransmissions): %9li %9li %9li\n",
         sim->stat->read[0],sim->stat->read[1],sim->stat->read[2]);
   if (sim->stat->write[1]>0)   
      avg_w1time=((sim->stat->total_w1time.tv_sec+sim->stat->total_w1time.tv_usec/1000000.0)
                 *1000)/sim->stat->write[1];
   else
      avg_w1time=0;
   fprintf(sim->log_fd,"\nAverage total time for write with success in first attempt (ms): %9i\n%s\n",
         avg_w1time,line);      
      
   fclose(sim->log_fd);
   
   return 0;
}

    
void write_sigalrm_handler()
{  
   /*  SIGALRM is raised after the expiry of the run time by
       a time initialized in all write/read processes.
    */
    
   sim->f_stop_readwrite=1;
}   


void write_sigterm_handler()
{  
   /*  SIGTERM is raised after one of the write or read
       processes exits with a status != 0 or the 
       Stop or Exit button is pressed.
    */
    
   shmdt(sim->shm_results);
   if (rw_sim.eib_fd>0)
      close(rw_sim.eib_fd);
   exit(97);
}   


void sim_write(Eib_bcu2_sim_shm *shm_segment)
{  int flags;
   unsigned long t1,t2,tu,tp,tp1;
   clock_t start_time,end_time;
   struct tms tms_start,tms_end;
   ssize_t nwritten;
   int msg_len;
   double ms_per_clock;
   int i;
   struct sigaction write_sigalrm;
   struct sigaction write_sigterm;


   char *msg="ZBCDEFGHIJKLMNOPQRSTUVW";  // dummy message written to the driver
//   char *msg=(char *)malloc(23*sizeof(char)); // for PEI_IDENTIFY-service
//   msg[0]=0xA7; 

   // must be before P()
   sim->f_stop_readwrite=0;
   rw_sim.eib_fd=0;
   
   P(sim->sem_start_rw);     // wait for start of all write and read processes
                             // unlocked by split process after creation of 
                             // all write and read processes

   rw_sim.shm_segment=shm_segment;

   rw_sim.local_results.result=0;
   rw_sim.local_results.tp=0;
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++) 
   {  rw_sim.local_results.rw[i]=0;
      rw_sim.local_results.w_failed[i]=0;
   }
   write_sigalrm.sa_handler=write_sigalrm_handler;
   sigfillset(&write_sigalrm.sa_mask);
   write_sigalrm.sa_flags=0;
   if (sigaction(SIGALRM,&write_sigalrm,NULL)==-1)
   {  shmdt(sim->shm_results);
      exit(10);
   }
   write_sigterm.sa_handler=write_sigterm_handler;
   sigfillset(&write_sigterm.sa_mask);
   write_sigterm.sa_flags=0;
   if (sigaction(SIGTERM,&write_sigterm,NULL)==-1)
   {  shmdt(sim->shm_results);
      exit(11);
   }
   
   switch(sim->sim_parameter->wmode)
   {  case EIBDRV_SIM_NONBLOCKSTDWMODE: flags=O_RDWR|O_NONBLOCK; break;
      case EIBDRV_SIM_BLOCKSTDWMODE:    flags=O_RDWR; break;
      case EIBDRV_SIM_NONBLOCKACKWMODE: flags=O_RDWR|O_NONBLOCK; break;
      case EIBDRV_SIM_BLOCKACKWMODE:    flags=O_RDWR; break;
   }
   
   if ((rw_sim.eib_fd=open(sim->device,flags,0))<0)
   {  shmdt(sim->shm_results);
      exit(12);
   }
   
   if ((sim->sim_parameter->wmode==EIBDRV_SIM_BLOCKACKWMODE)&&
       (sim->ftstation_settings->mode==FT_SERVERMODE))
   {  if (ioctl(rw_sim.eib_fd,FT_SET_WRITE_MODE,FT_WRITE_ACKNOWLEDGED)<0)
      {  shmdt(sim->shm_results);
         close(rw_sim.eib_fd);
         exit(13);
      }
   }

   /*  if processes pass that point within one second, same random sequences
       are createc because of same time argument; use getpid() to avoid this.
    */  
    
   srand(time(NULL)-getpid());             

   alarm(sim->sim_parameter->run_time*60);
   ms_per_clock=1000/sysconf(_SC_CLK_TCK);
   tp=0;tp1=0;

   while(!sim->f_stop_readwrite)
   {  t1=rand() % sim->sim_parameter->write_interval;
      if (sim->sim_parameter->winterval==EIBDRV_SIM_REDINTERVAL)
      {  if (tp>t1) {  tp-=t1; tp1=t1; t1=0;  }
         else       {  tp1=tp; t1-=tp; tp=0;  }
      }
      if ((sim->sim_parameter->max_len - sim->sim_parameter->min_len)>0)
         msg_len=(rand() % (sim->sim_parameter->max_len - sim->sim_parameter->min_len + 1))+ 
            sim->sim_parameter->min_len;
      else
         msg_len=sim->sim_parameter->min_len;
      delay(t1*1000);
      start_time=times(&tms_start);
      if ((nwritten=write(rw_sim.eib_fd,msg,msg_len))==msg_len)
         rw_sim.local_results.rw[msg_len-1]++;
      else
         rw_sim.local_results.w_failed[msg_len-1]++;
      end_time=times(&tms_end);
      if ((tu=t1+tp1+(time_t)((end_time-start_time)*ms_per_clock))<sim->sim_parameter->write_interval)
      {  t2=sim->sim_parameter->write_interval-tu;
         if (sim->sim_parameter->winterval==EIBDRV_SIM_REDINTERVAL)
         {  if (tp>t2) {  tp-=t2; t2=0;  }
            else       {  t2-=tp; tp=0;  }
         }
      } 
      else
      {  t2=0;
         if (sim->sim_parameter->winterval==EIBDRV_SIM_REDINTERVAL)
            tp+=tu-sim->sim_parameter->write_interval;
            
      }
      delay(t2*1000);
      rw_sim.local_results.tp=tp;
   }
   alarm(0);  
   
   /* copy local results into shared memory segment */
   P(sim->sem_shm_results);
   *rw_sim.shm_segment=rw_sim.local_results;
   V(sim->sem_shm_results);

   shmdt(sim->shm_results);
   close(rw_sim.eib_fd);
   exit(0);
}


void read_sigio_handler()
{  
   /*  SIGIO is raised, when a message is available at the eibdrv driver. If
       the driveris operated in Standard Mode and several read processes are
       created, they compete for messages and my return with result 0.
    */
    
   ssize_t nread;
   unsigned char msg[EIB_BUF_ELEMENT_SIZE];
   
   if ((nread=read(rw_sim.eib_fd,&msg[0],EIB_BUF_ELEMENT_SIZE))>0)
      rw_sim.local_results.rw[nread-1]++;
}


void read_sigalrm_handler()
{  
   /*  SIGALRM is raised after the expiry of the run time by
       a time initialized in all write/read processes.
    */
    
   sim->f_stop_readwrite=1;
}


void read_sigterm_handler()
{
   /*  SIGTERM is raised after one of the write or read
       processes exits with a status != 0 or the 
       Stop or Exit button is pressed.
    */

   shmdt(sim->shm_results);
   if (rw_sim.eib_fd>0)
      close(rw_sim.eib_fd);
   exit(97);
}


void sim_read(Eib_bcu2_sim_shm *shm_segment)
{  int i;
   unsigned long flags;
   struct sigaction read_sigalrm;
   struct sigaction read_sigterm;
   struct sigaction read_sigio;

   // must be before P()
   sim->f_stop_readwrite=0;
   rw_sim.eib_fd=0;
         
   P(sim->sem_start_rw);
   rw_sim.shm_segment=shm_segment;

   rw_sim.local_results.result=0;
   rw_sim.local_results.tp=0;
   for (i=0;i<EIB_BUF_ELEMENT_SIZE;i++) 
   {  rw_sim.local_results.rw[i]=0;
      rw_sim.local_results.w_failed[i]=0;
   }

   read_sigalrm.sa_handler=read_sigalrm_handler;
   sigfillset(&read_sigalrm.sa_mask);
   read_sigalrm.sa_flags=0;
   if (sigaction(SIGALRM,&read_sigalrm,NULL)==-1)
   {  shmdt(sim->shm_results);
      exit(20);
   }

   read_sigterm.sa_handler=read_sigterm_handler;
   sigfillset(&read_sigterm.sa_mask);
   read_sigterm.sa_flags=0;
   if (sigaction(SIGTERM,&read_sigterm,NULL)==-1)
   {  shmdt(sim->shm_results);
      exit(21);
   }

   read_sigio.sa_handler=read_sigio_handler;
   sigfillset(&read_sigio.sa_mask);
   read_sigio.sa_flags=0;
   if (sigaction(SIGIO,&read_sigio,NULL)==-1)
   {  shmdt(sim->shm_results);
      exit(22);
   }
   
   if ((rw_sim.eib_fd=open(sim->device,O_RDWR|O_NONBLOCK))<0)
   {  shmdt(sim->shm_results);
      exit(23);
   }
   
   if ((fcntl(rw_sim.eib_fd,F_SETOWN,getpid())==-1)||
       ((flags=fcntl(rw_sim.eib_fd,F_GETFL))==-1) ||
       ((fcntl(rw_sim.eib_fd,F_SETFL,flags|FASYNC)==-1)))
   {  shmdt(sim->shm_results);
      exit(24);
   }

   alarm(sim->sim_parameter->run_time*60);      
   while(!sim->f_stop_readwrite) 
      pause();

   alarm(0);   

   /* Copy local results into shared memory segment */
   P(sim->sem_shm_results);
   *rw_sim.shm_segment=rw_sim.local_results;
   V(sim->sem_shm_results);

   shmdt(sim->shm_results);
   close(rw_sim.eib_fd);
   exit(0);
}


void split_sigterm_handler()
{  
   /*  SIGTERM is raised after one of the write or read
       processes exits with a status != 0 or the 
       Stop or Exit button is pressed.
       
       This signal handler is inherited to write and read processes
       until they register their own SIGTERM-Handlers.
       
       The split process never terminates upper the SIGTERM-Signal;
       this handler is necessary because of sending the SIGTERM
       signal to the whole process group of which split is 
       process group leader.
    */
    
   if (getpgrp()!=getpid())
   {   shmdt(sim->shm_results);
       exit(97);
   }
}


void sim_split()
{  int i,j,result;
   struct sigaction split_sigchld;
   struct sigaction split_sigterm;
   pid_t exit_pid;
   int exit_status;
   int exit_code=0;

   split_sigterm.sa_handler=split_sigterm_handler;
   sigfillset(&split_sigterm.sa_mask);
   split_sigterm.sa_flags=0;
   if (sigaction(SIGTERM,&split_sigterm,NULL)==-1)
      exit(3);

   split_sigchld.sa_handler=SIG_DFL;
   sigfillset(&split_sigchld.sa_mask);
   split_sigchld.sa_flags=0;
   if (sigaction(SIGCHLD,&split_sigchld,NULL)==-1)
      exit(4);
      
   /*  Create a process group and make itself to its leader */
   if (setpgid(getpid(),0)!=0)
      exit(5);

   /*  Fork write and read processes */ 
   for (j=0;j<2;j++)
   {  for (i=0;i<sim->sim_parameter->proc_count;i++)
      {  result=fork();
         if (result==0)
         {  if (j==0) sim_write(sim->shm_results+j*sim->sim_parameter->proc_count+i); 
            else      sim_read(sim->shm_results+j*sim->sim_parameter->proc_count+i);  
         }
         else if (result==-1)
            break;
      }
      if (result==-1)
         break;
   }

   /*  In case of error terminate all processes created */
   if (result==-1)
   {  kill(0,SIGTERM);
      exit_code=1;
   }
   
   /*  Unlock all read and write processes by a global semaphore variable */
   if (semctl(sim->sem_start_rw,0,SETVAL,(int)2*sim->sim_parameter->proc_count)==-1)
      exit_code=2;
   
   /* Wait for termination of all write/read processes */
   for (i=0;i<sim->sim_parameter->proc_count*2;i++)
   {  exit_pid=wait(&exit_status);
      if ((WIFSIGNALED(exit_status))||
          ((WIFEXITED(exit_status))&&(WEXITSTATUS(exit_status)>0)))
      {  kill(0,SIGTERM); 
         if (exit_code==0) 
            exit_code=WEXITSTATUS(exit_status);
      }
   }
   
   exit(exit_code);
}



void cleanup(void *arg,int btn)
{
   /*  Return all semaphores, shared memory segments; close all open
       files; hide all forms and return the central simulator struct.
       
       This function is called in case of error during process creation
       in main function.
    */
   
   Eib_bcu2_sim *sim=(Eib_bcu2_sim *)arg;
   
   switch(sim->cleanup_code)
   {  
      case 8:  shmdt(sim->shm_results);
      case 7:  shmctl(sim->shm_results_id,IPC_RMID,(int)0);
      case 6:  fl_ext_activate_object(sim->MainForm->StartBtn);
               return;

      case 5:  semctl(sim->sem_start_rw,0,IPC_RMID,(int)0);
      case 4:  semctl(sim->sem_shm_results,0,IPC_RMID,(int)0);
      case 3:  fl_free_form(sim->MainForm->VisForm);
               fl_ext_free_indicator(sim->RunForm->Indicator);
               fl_free_form(sim->RunForm->VisForm);
      case 2:  close(sim->eib_fd);
      case 1:  returnEib_bcu2_sim(sim);
               fl_finish();
               exit(3);
   }
}

void main_sigchld_handler()
{  
   /*  SIGCHLD is raised, when the split process terminates after the termination
       of all write/read processes upon the expiry of runtime or termination 
       by Stop or Exit Btn.
    */
    
   FD_ResultsForm *ResultsForm;
   int exit_status;
   pid_t exit_pid;
   int result;
   
   if (sim->state==EIBDRV_SIM_STATE_RUNNING)
   {  sim->state=EIBDRV_SIM_STATE_READY;
      exit_pid=wait(&exit_status);
      if (WIFSIGNALED(exit_status))
      {  make_InternalErrorPanel(sim->Panels+6,98);
         fl_ext_panel1(sim->Panels+6,sim,panel_cb);
      }
      else if ((WIFEXITED(exit_status))&&(WEXITSTATUS(exit_status)>0))
      {  switch(WEXITSTATUS(exit_status))
         {  case 97: fl_ext_panel1(sim->Panels+5,sim,panel_cb);
                     break;
            default: make_InternalErrorPanel(sim->Panels+6,WEXITSTATUS(exit_status));
                     fl_ext_panel1(sim->Panels+6,sim,panel_cb);
                     break;
         }
      }
      else
      {  
         time(&sim->stat_get_time);  
         if (ioctl(sim->eib_fd,FT_GET_STAT,sim->stat)<0)
         {  switch(result=errno)
            {  case EFAULT: make_InternalErrorPanel(sim->Panels+6,59);
                            fl_ext_panel1(sim->Panels+6,sim,panel_cb);
                            break;
               case EBUSY:  fl_ext_panel1(sim->Panels+4,sim,panel_cb);
                            break;
               case EACCES: fl_ext_panel1(sim->Panels+10,sim,panel_cb);
                            break;
            }
         }
         else
         {  ResultsForm=create_form_ResultsForm(sim);
            fl_show_form(ResultsForm->VisForm,FL_PLACE_CENTER,FL_TRANSIENT,"eibdrv_sim");
            if (create_Log(sim)<0)   
               fl_ext_panel1(sim->Panels+11,sim,panel_cb);
         }
      }
     
      shmdt(sim->shm_results);
      shmctl(sim->shm_results_id,IPC_RMID,(int)0);

      fl_ext_activate_object(sim->MainForm->StartBtn);
      fl_hide_form(sim->RunForm->VisForm);
#if FL_INCLUDE_VERSION < 89      
      fl_set_idle_callback(0,0);
#endif
   }
   else // sim->state==EIB_BCU_SIM_STATE_SHUTDOWN
   {  shmdt(sim->shm_results);
      shmctl(sim->shm_results_id,IPC_RMID,(int)0);
      semctl(sim->sem_start_rw,0,IPC_RMID,(int)0);
      semctl(sim->sem_shm_results,0,IPC_RMID,(int)0);
      fl_free_form(sim->MainForm->VisForm);
      fl_ext_free_indicator(sim->RunForm->Indicator);
      fl_free_form(sim->RunForm->VisForm);
      close(sim->eib_fd);
      returnEib_bcu2_sim(sim);
      fl_finish();
      exit(0);
   }
   
}


int main(int argc,char *argv[])
{  int result;
   int len;
   struct sigaction main_sigchld;
         
   sim=getEib_bcu2_sim();

   if (read_PanelContentA(EIBDRV_SIM_PANELFILE,&sim->Panels)!=EIBDRV_SIM_PANELFILESIZE)
   {  printf("eibdrv_sim: Panelfile %s not found!\n",EIBDRV_SIM_PANELFILE);
      exit(2);
   }

   if (argc!=2)
   {  printf("usage: eibdrv_sim device\n");
      exit(1);
   }

   sim->device=(char *)malloc((strlen(argv[1])+1)*sizeof(char));
   strcpy(sim->device,argv[1]);
   
   fl_set_icm_color(FL_INACTIVE_COL,200,200,200);
   fl_initialize(&argc,argv,"eibdrv_sim",0,0);
   fl_set_border_width(2);

   sim->cleanup_code=1;
      
   if ((sim->eib_fd=open(argv[1],O_RDWR|O_NONBLOCK))<0)
   {  switch(result=errno)
      {  case ENOENT:  fl_ext_panel1(sim->Panels,sim,cleanup);
                       break;
         case EBUSY:   fl_ext_panel1(sim->Panels+1,sim,cleanup);
                       break;
         case ENOBUFS: fl_ext_panel1(sim->Panels+2,sim,cleanup);
                       break;
         default:      fl_ext_panel1(sim->Panels+3,sim,cleanup);
                       break;
      }
      fl_do_forms();
   }
   sim->cleanup_code=2;

   if (((sim->ftstation_settings->mode=ioctl(sim->eib_fd,FT_GET_MODE))<0) ||
       ((sim->ftstation_settings->symb_baudrate=ioctl(sim->eib_fd,FT_GET_BAUDRATE))<0) ||
       ((len=ioctl(sim->eib_fd,FT_GET_PORT_LENGTH))<0) ||
       (ioctl(sim->eib_fd,FT_GET_OUTBUF_SIZE,&(sim->ftstation_settings->outbuf_size))<0) ||
       (ioctl(sim->eib_fd,FT_GET_WPBUF_SIZE,&(sim->ftstation_settings->wpbuf_size))<0) ||
       (ioctl(sim->eib_fd,FT_GET_RPBUF_SIZE,&(sim->ftstation_settings->rpbuf_size))<0) ||
       (ioctl(sim->eib_fd,FT_GET_RPBUF_MSGSIZE,&(sim->ftstation_settings->rpbuf_msgbuf_size))<0))
   {  switch(result=errno)
      {  case EBUSY:  fl_ext_panel1(sim->Panels+4,sim,cleanup);
                      fl_do_forms();
         case EFAULT: fl_ext_panel1(sim->Panels+6,sim,cleanup);
                      fl_do_forms();
         case EACCES: fl_ext_panel1(sim->Panels+10,sim,cleanup);
                      fl_do_forms();
         case EINVAL: sim->ftstation_settings->outbuf_size=0;
                      sim->ftstation_settings->wpbuf_size=0;
                      sim->ftstation_settings->rpbuf_size=0;
                      sim->ftstation_settings->rpbuf_msgbuf_size=0; 
                      break;
      }
   }
   
   sim->ftstation_settings->port=(char *)malloc((len+1)*sizeof(char));
   if (ioctl(sim->eib_fd,FT_GET_PORT,sim->ftstation_settings->port)<0)
   {  switch(result=errno)
      {  case EFAULT: make_InternalErrorPanel(sim->Panels+6,50);
                      fl_ext_panel1(sim->Panels+6,sim,cleanup);
                      break;
         case EBUSY:  fl_ext_panel1(sim->Panels+4,sim,cleanup);
                      break;
         case EACCES: fl_ext_panel1(sim->Panels+10,sim,cleanup);
                      break;
      }
      fl_do_forms();
   }
   
   sim->MainForm=create_form_MainForm(sim);
   sim->RunForm=create_form_RunForm(sim);
   fl_set_object_label(sim->MainForm->DeviceText,argv[1]);
   switch(sim->ftstation_settings->mode)
   {  case FT_STANDARDMODE: 
              fl_set_object_label(sim->MainForm->ModeText,"Standard Mode");
              fl_ext_deactivate_object(sim->MainForm->NonblockingWModeCheckbtn);
              fl_ext_deactivate_object(sim->MainForm->BlockingWModeCheckbtn);
              break;
      case FT_SERVERMODE:
              fl_set_object_label(sim->MainForm->ModeText,"Server Mode");
              break;
   }

   sim->cleanup_code=3;

   if ((sim->sem_shm_results=semget(IPC_PRIVATE,1,0600|IPC_CREAT|IPC_EXCL))==-1)
   {  make_InternalErrorPanel(sim->Panels+6,51);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      fl_do_forms();
   }
   sim->cleanup_code=4;
   
   if (semctl(sim->sem_shm_results,0,SETVAL,(int)1)==-1)
   {  make_InternalErrorPanel(sim->Panels+6,52);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      fl_do_forms();
   }

   if ((sim->sem_start_rw=semget(IPC_PRIVATE,1,0600|IPC_CREAT|IPC_EXCL))==-1)
   {  make_InternalErrorPanel(sim->Panels+6,53);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      fl_do_forms();
   }
   sim->cleanup_code=5;
   
   if (semctl(sim->sem_start_rw,0,SETVAL,(int)0)==-1)
   {  make_InternalErrorPanel(sim->Panels+6,54);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      fl_do_forms();
   }

   main_sigchld.sa_handler=main_sigchld_handler;
   sigfillset(&main_sigchld.sa_mask);
   main_sigchld.sa_flags=0;
   if (sigaction(SIGCHLD,&main_sigchld,NULL)==-1)
   {  make_InternalErrorPanel(sim->Panels+6,55);
      fl_ext_panel1(sim->Panels+6,sim,cleanup);
      fl_do_forms();
   }

   fl_show_form(sim->MainForm->VisForm,FL_PLACE_CENTER,FL_TRANSIENT,"eibdrv_sim");

   fl_do_forms();
   exit(0);
}
